"""
Drift analysis for localization coordinates.

This module provides functions for estimating spatial drift in localization data.

Software based drift correction using image correlation has been described in several publications .
Methods employed for drift estimation comprise single molecule localization analysis (an iterative closest point (icp)
algorithm as implemented in the open3d library [1]_, [2]_) or image cross-correlation analysis [3]_, [4]_, [5]_.

Examples
--------
Please use the following procedure to estimate and correct for spatial drift::

    from lmfit import LinearModel
    drift = Drift(chunk_size=1000, target='first').\\
            compute(locdata).\\
            fit_transformations(slice_data=slice(0, -1),
                                matrix_models=None,
                                offset_models=(LinearModel(), LinearModel())).\\
            apply_correction()
    locdata_corrected = drift.locdata_corrected

References
----------
.. [1] Qian-Yi Zhou, Jaesik Park, Vladlen Koltun,
   Open3D: A Modern Library for 3D Data Processing,
   arXiv 2018, 1801.09847

.. [2] Rusinkiewicz and M. Levoy,
   Efficient variants of the ICP algorithm,
   In 3-D Digital Imaging and Modeling, 2001.

.. [3] C. Geisler,
   Drift estimation for single marker switching based imaging schemes,
   Optics Express. 2012, 20(7):7274-89.

.. [4] Yina Wang et al.,
   Localization events-based sample drift correction for localization microscopy with redundant cross-correlation
   algorithm, Optics Express 2014, 22(13):15982-91.

.. [5] Michael J. Mlodzianoski et al.,
   Sample drift correction in 3D fluorescence photoactivation localization microscopy,
   Opt Express. 2011 Aug 1;19(16):15009-19.

"""
import sys
from collections import namedtuple
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.interpolate import splev, splrep
from lmfit.models import ConstantModel, LinearModel, PolynomialModel

from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.data.locdata import LocData
from locan.data.register import _register_icp_open3d, register_cc
from locan.data.transform.transformation import transform_affine
from locan.data.metadata_utils import _modify_meta

if HAS_DEPENDENCY["open3d"]: import open3d as o3d


__all__ = ['Drift', 'DriftComponent']

logger = logging.getLogger(__name__)


##### The algorithms

@needs_package("open3d")
def _estimate_drift_icp(locdata, chunks=None, chunk_size=None, target='first', kwargs_chunk=None, kwargs_register=None):
    """
    Estimate drift from localization coordinates by registering points in successive time-chunks of localization
    data using an "Iterative Closest Point" algorithm.

    Parameters
    ----------
    locdata : LocData
       Localization data with properties for coordinates and frame.
    chunks : list[tuples]
        Localization chunks as defined by a list of index-tuples
    chunk_size : int
       Number of consecutive localizations to form a single chunk of data.
    target : str
       The chunk on which all other chunks are aligned. One of 'first', 'previous'.
    kwargs_chunk : dict
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_register : dict
        Other parameter passed to :func:`_register_icp_open3d`.

    Returns
    -------
    tuple(LocData, list of namedtuple('Transformation', 'matrix offset'))
        Collection and corresponding transformations.
    """
    if kwargs_chunk is None:
        kwargs_chunk = {}
    if kwargs_register is None:
        kwargs_register = {}

    # split in chunks
    collection = LocData.from_chunks(locdata, chunks=chunks, chunk_size=chunk_size, **kwargs_chunk)

    # register locdatas
    # initialize with identity transformation for chunk zero.
    Transformation = namedtuple('Transformation', 'matrix offset')
    transformations = [Transformation(np.identity(locdata.dimension), np.zeros(locdata.dimension))]
    if target == 'first':
        for locdata in collection.references[1:]:
            transformation = _register_icp_open3d(locdata.coordinates, collection.references[0].coordinates,
                                                  **dict(dict(matrix=None, offset=None, pre_translation=None,
                                                  max_correspondence_distance=100, max_iteration=10_000,
                                                  verbose=False), **kwargs_register))
            transformations.append(transformation)

    elif target == 'previous':
        for n in range(len(collection.references)-1):
            transformation = _register_icp_open3d(collection.references[n+1].coordinates,
                                                  collection.references[n].coordinates,
                                                  **dict(dict(matrix=None, offset=None, pre_translation=None,
                                                  max_correspondence_distance=100, max_iteration=10_000,
                                                  with_scaling=False, verbose=False), **kwargs_register))
            transformations.append(transformation)

    return collection, transformations


def _estimate_drift_cc(locdata, chunks=None, chunk_size=None, target='first', bin_size=10,
                       kwargs_chunk=None, kwargs_register=None):
    """
    Estimate drift from localization coordinates by registering points in successive time-chunks of localization
    data using a cross-correlation algorithm.

    Parameters
    ----------
    locdata : LocData
       Localization data with properties for coordinates and frame.
    chunks : list[tuples]
        Localization chunks as defined by a list of index-tuples
    chunk_size : int
       Number of consecutive localizations to form a single chunk of data.
    target : str
       The chunk on which all other chunks are aligned. One of 'first', 'previous'.
    bin_size : tuple of int, float
        Size per image pixel
    kwargs_chunk : dict
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_register : dict
        Other parameter passed to :func:`register_cc`.

    Returns
    -------
    Locdata and list of namedtuple
        collection and corresponding transformations.
    """
    if kwargs_chunk is None:
        kwargs_chunk = {}
    if kwargs_register is None:
        kwargs_register = {}

    # split in chunks
    collection = LocData.from_chunks(locdata, chunks=chunks, chunk_size=chunk_size, **kwargs_chunk)
    ranges = locdata.bounding_box.hull.T

    # register images
    # initialize with identity transformation for chunk zero.
    Transformation = namedtuple('Transformation', 'matrix offset')
    transformations = [Transformation(np.identity(locdata.dimension), np.zeros(locdata.dimension))]
    if target == 'first':
        for reference in collection.references[1:]:
            transformation = register_cc(reference, collection.references[0],
                                         **dict(dict(range=ranges, bin_size=bin_size), **kwargs_register))
            transformations.append(transformation)

    elif target == 'previous':
        for n in range(len(collection) - 1):
            transformation = register_cc(collection.references[n + 1], collection.references[n],
                                         **dict(dict(range=ranges, bin_size=bin_size), **kwargs_register))
            transformations.append(transformation)

    return collection, transformations


##### The specific analysis classes

class _LmfitModelFacade:

    def __init__(self, model):
        self.model = model
        self.model_result = None

    def fit(self, x, y, verbose=False, **kwargs):
        params = self.model.guess(data=y, x=x)
        self.model_result = self.model.fit(x=x, data=y, params=params, **kwargs)
        if verbose:
            self.plot()
        return self.model_result

    def eval(self, x):
        x = np.asarray(x)
        return self.model_result.eval(x=x)

    def plot(self, **kwargs):
        return self.model_result.plot(**kwargs)


class _ConstantModelFacade:

    def __init__(self, **kwargs):
        self.model = ConstantModel(**kwargs)
        self.model_result = None
        self.independent_variable = None

    def fit(self, x, y, verbose=False, **kwargs):
        self.independent_variable = x
        self.model_result = self.model.fit(x=x, data=y, **kwargs)
        if verbose:
            self.plot()
        return self.model_result

    def eval(self, x):
        if isinstance(x, (tuple, list, np.ndarray)):
            return np.array([self.model_result.eval()] * len(x))
        else:
            return self.model_result.eval(x=x)

    def plot(self, **kwargs):
        x = self.independent_variable
        y = self.model_result.data
        return plt.plot(x, y, 'o', x, self.eval(x=x), **kwargs)


class _ConstantZeroModelFacade(_ConstantModelFacade):

    def __init__(self):
        self.model = ConstantModel()
        self.model.set_param_hint(name='c', value=0, vary=False)
        self.model_result = None


class _ConstantOneModelFacade(_ConstantModelFacade):

    def __init__(self):
        self.model = ConstantModel()
        self.model.set_param_hint(name='c', value=1, vary=False)
        self.model_result = None


class _SplineModelFacade:

    def __init__(self, **kwargs):
        self.model = 'spline'
        self.model_result = None
        self.parameter = kwargs
        self.independent_variable = None
        self.data = None

    def fit(self, x, y, verbose=False, **kwargs):
        self.independent_variable = x
        self.data = y
        self.model_result = splrep(x, y, **dict(dict(k=3, s=100), **dict(**self.parameter, **kwargs)))
        if verbose:
            self.plot()
        return self.model_result

    def eval(self, x):
        np.asarray(x)
        results = splev(x, self.model_result)
        if isinstance(x, (tuple, list, np.ndarray)):
            return results
        else:
            return float(results)

    def plot(self, **kwargs):
        x = self.independent_variable
        y = self.data
        x_ = np.linspace(np.min(x), np.max(x), 100)
        return plt.plot(x, y, 'o', x_, self.eval(x=x_), **kwargs)


class DriftComponent:
    """
    Class carrying model functions to describe drift over time (in unit of frames). DriftComponent provides
    a transformation to apply a drift correction.

    Standard models for constant, linear or polynomial drift correction are taken from :mod:`lmfit.models`.
    For fitting splines we use the scipy function :func:`scipy.interpolate.splrep`.

    Parameters
    ----------
    type : str, lmfit.models.Model, None
        Model class or indicator for setting up the corresponding model class.
        String can be one of `zero`, `one`, `constant`, `linear`, `polynomial`, `spline`.

    Attributes
    ----------
    type : str
        String indicator for model.
    model : lmfit.models.Model, return value of :func:`scipy.interpolate.splrep`
        The model definition
    model_result : lmfit.model.ModelResult, collection of model results.
        The results collected from fitting the model to specified data.
    """

    def __init__(self, type=None, **kwargs):
        self.type = type
        self.model_result = None

        if type is None:
            self.type = 'none'
            self.model = None
        elif type == 'zero':
            self.model = _ConstantZeroModelFacade()
        elif type == 'one':
            self.model = _ConstantOneModelFacade()
        elif type == 'constant':
            self.model = _ConstantModelFacade(**kwargs)
        elif type == 'linear':
            self.model = _LmfitModelFacade(LinearModel(**kwargs))
        elif type == 'polynomial':
            self.model = _LmfitModelFacade(PolynomialModel(**dict(dict(degree=3), **kwargs)))
        elif getattr(type, '__module__', None) == 'lmfit.models':
            self.type = type.name
            self.model = _LmfitModelFacade(model=type)
        elif type == 'spline':
            self.model = _SplineModelFacade(**kwargs)
        else:
            raise TypeError(f"DriftComponent cannot handle type={type}.")

    def fit(self, x, y, verbose=False, **kwargs):
        """
        Fit model to the given data and create `self.model_results`.

        Parameters
        ----------
        x : array-like
            x data
        y : array-like
            y values
        verbose : bool
            show plot
        kwargs : dict
            Other parameters passed to :func:`lmfit.model.fit` or to :func:`scipy.interpolate.splrep`
            Use the parameter `s` to set the amount of smoothing.

        Returns
        -------
        self
        """
        self.model_result = self.model.fit(x, y, verbose=verbose, **kwargs)
        return self

    def eval(self, x):
        """
        Compute a transformation for time `x` from the drift model.

        Parameters
        ----------
        x : array-like
            frame values

        Returns
        -------
        Array of float.
        """
        return self.model.eval(x)


class Drift(_Analysis):
    """
    Estimate drift from localization coordinates by registering points in successive time-chunks of localization
    data using an iterative closest point algorithm (icp) or image cross-correlation algorithm (cc).

    Parameters
    ----------
    locdata : LocData
        Localization data representing the source on which to perform the manipulation.
    chunks : list[tuples]
        Localization chunks as defined by a list of index-tuples
    chunk_size : int
        Number of consecutive localizations to form a single chunk of data.
    target : str
        The chunk on which all other chunks are aligned. One of 'first', 'previous'.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    method : str
        The method used for computation.
        One of iterative closest point algorithm 'icp' or image cross-correlation algorithm 'cc'.
    bin_size : tuple of int, float
        Only for method='cc': Size per image pixel
    kwargs_chunk : dict
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_icp : dict
        Other parameter passed to :func:`_register_icp_open3d`.
    kwargs_cc : dict
        Other parameter passed to :func:`register_cc`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    locdata : LocData
        Localization data representing the source on which to perform the manipulation.
    collection : Locdata object
        Collection of locdata chunks
    transformations : list of namedtuple('Transformation', 'matrix offset')
        Transformations for locdata chunks
    transformation_models : namedtuple('Transformation_models', 'matrix offset') with lists of lmfit.ModelResult objects.
        The fitted model objects.
    locdata_corrected : LocData
        Localization data with drift-corrected coordinates.
    """
    count = 0

    def __init__(self, meta=None, chunks=None, chunk_size=None, target='first', method='icp',
                 kwargs_chunk=None, kwargs_register=None):
        super().__init__(meta, chunks=chunks, chunk_size=chunk_size, target=target, method=method,
                         kwargs_chunk=kwargs_chunk, kwargs_register=kwargs_register)
        self.locdata = None
        self.collection = None
        self.transformations = None
        self.transformation_models = dict(matrix=None, offset=None)
        self.locdata_corrected = None

    def __bool__(self):
        if self.transformations is not None:
            return True
        else:
            return False

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data representing the source on which to perform the manipulation.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        if self.parameter['method'] == 'icp':
            collection, transformations = _estimate_drift_icp(locdata,
                                                              chunks=self.parameter['chunks'],
                                                              chunk_size=self.parameter['chunk_size'],
                                                              kwargs_chunk=self.parameter['kwargs_chunk'],
                                                              kwargs_register=self.parameter['kwargs_register'])
        elif self.parameter['method'] == 'cc':
            collection, transformations = _estimate_drift_cc(locdata,
                                                             chunks=self.parameter['chunks'],
                                                             chunk_size=self.parameter['chunk_size'],
                                                             kwargs_chunk=self.parameter['kwargs_chunk'],
                                                             kwargs_register=self.parameter['kwargs_register'])
        else:
            raise ValueError(f'Method {self.parameter["method"]} is not defined. One of "icp", "cc".')
        self.locdata = locdata
        self.collection = collection
        self.transformations = transformations
        self.transformation_models = dict(matrix=None, offset=None)
        self.locdata_corrected = None
        return self

    def _transformation_models_for_identity_matrix(self):
        """Return transformation_models (dict) with DriftModels according to unit matrix."""
        dimension = self.locdata.dimension
        transformation_models = []
        for k in np.identity(dimension).flatten():
            if k == 0:
                transformation_models.append(DriftComponent('zero'))
            else:  # if k == 1
                transformation_models.append(DriftComponent('one'))
        return dict(matrix=transformation_models)

    def _transformation_models_for_zero_offset(self):
        """Return transformation_models (dict) with DriftModels according to zero offset."""
        dimension = self.locdata.dimension
        return dict(offset=[DriftComponent('zero') for _ in range(dimension)])

    def fit_transformation(self, slice_data=slice(None), transformation_component='matrix', element=0,
                           drift_model='linear', verbose=False):
        """
        Fit drift model to selected component of the estimated transformations.

        Parameters
        ----------
        slice_data : Slice object
            Reduce data to a selection on the localization chunks.
        transformation_component : str
            One of 'matrix' or 'offset'
        element : int, None
            The element of flattened transformation matrix or offset
        drift_model : DriftComponent, str, None
            A drift model as defined by a :class:`DriftComponent` instance
            or the parameter `type` as defined in :class:`DriftComponent`.
            For None no change will occur. To reset transformation_models set the transformation_component to None:
            e.g. self.transformation_components = None or self.transformation_components['matrix'] = None.
        verbose : bool
            Print the fitted model curves using lmfit.
        """
        if not self:
            logger.warning('No transformations available to be fitted.')
            return self

        dimension = self.locdata.dimension
        # frames = np.array([locdata.data.frame.mean() for locdata in self.collection.references[1:]])[slice_data]
        frames = np.array([locdata.data.frame.mean() for locdata in self.collection.references])[slice_data]

        if drift_model is None:
            return self

        if not isinstance(drift_model, DriftComponent):
            drift_model = DriftComponent(type=drift_model)

        if self.parameter['target'] == 'first':
            if transformation_component == 'matrix':
                y = np.array([transformation.matrix[element // dimension][element % dimension]
                              for transformation in self.transformations])[slice_data]
                if self.transformation_models['matrix'] is None:
                    self.transformation_models.update(self._transformation_models_for_identity_matrix())
            elif transformation_component == 'offset':
                y = np.array([transformation.offset[element] for transformation in self.transformations])[slice_data]
                if self.transformation_models['offset'] is None:
                    self.transformation_models.update(self._transformation_models_for_zero_offset())
            else:
                raise ValueError("transformation_component must be 'matrix' or 'offset'.")

            self.transformation_models[transformation_component][element] = drift_model
            self.transformation_models[transformation_component][element].fit(frames, y, verbose=False)

            if verbose:
                self.transformation_models[transformation_component][element].model.plot()

        elif self.parameter['target'] == 'previous':
            raise NotImplementedError("Not yet implemented. Use target = 'first'.")

        return self

    def fit_transformations(self, slice_data=slice(None), matrix_models=None, offset_models=None,
                            verbose=False):
        """
        Fit model parameter to all estimated transformation components.

        Parameters
        ----------
        slice_data : Slice object
            Reduce data to a selection on the localization chunks.
        matrix_models : list of DriftComponent, None
            Models to use for fitting each matrix component.
            Length of list must be equal to the square of the transformations dimension (4 or 9).
            If None, no matrix transformation will be carried out when calling :func:`apply_correction`.
        offset_models : list of DriftComponent, None
            Models to use for fitting each offset component.
            Length of list must be equal to the transformations dimension (2 or 3).
            If None, no offset transformation will be carried out when calling :func:`apply_correction`.
        verbose : bool
            Print the fitted model curves.
        """
        if not self:
            logger.warning('No transformations available to be fitted.')
            return self

        if matrix_models is None:
            self.transformation_models['matrix'] = None
        else:
            if len(matrix_models) != self.locdata.dimension**2:
                raise ValueError('Length of matrix_models must be equal to the square of the '
                                 'transformations dimension (4 or 9).')
            self.transformation_models.update(self._transformation_models_for_identity_matrix())
            for n, matrix_model in enumerate(matrix_models):
                if matrix_model is None:
                    matrix_model = self.transformation_models['matrix'][n]
                self.fit_transformation(slice_data=slice_data, transformation_component='matrix', element=n,
                                        drift_model=matrix_model, verbose=verbose)

        if offset_models is None:
            self.transformation_models['offset'] = None
        else:
            if len(offset_models) != self.locdata.dimension:
                raise ValueError('Length of offset_models must be equal to the the transformations dimension (2 or 3).')
            self.transformation_models.update(self._transformation_models_for_zero_offset())
            for n, offset_model in enumerate(offset_models):
                if offset_model is None:
                    offset_model = self.transformation_models['offset'][n]
                self.fit_transformation(slice_data=slice_data, transformation_component='offset', element=n,
                                        drift_model=offset_model, verbose=verbose)

        return self

    def _apply_correction_on_chunks(self):
        """
        Correct drift by applying the estimated transformations to locdata chunks.
        """
        transformed_locdatas = []
        if self.parameter['target'] == 'first':
            transformed_locdatas = [transform_affine(locdata, transformation.matrix, transformation.offset)
                                    for locdata, transformation
                                    in zip(self.collection.references[1:], self.transformations)]

        elif self.parameter['target'] == 'previous':
            for n, locdata in enumerate(self.collection.references[1:]):
                transformed_locdata = locdata
                for transformation in reversed(self.transformations[:n]):
                    transformed_locdata = transform_affine(transformed_locdata,
                                                           transformation.matrix, transformation.offset)
                transformed_locdatas.append(transformed_locdata)

        new_locdata = LocData.concat([self.collection.references[0]] + transformed_locdatas)
        return new_locdata

    def _apply_correction_from_model(self, locdata):
        """
        Correct drift by applying the estimated transformations to locdata.
        If self.transformation_model['matrix'] is None, no matrix transformation will be carried out when calling
        :func:`apply_correction` (same for 'offset').

        Parameters
        ----------
        locdata : LocData or None
            Localization data to apply correction on. If None correction is applied to self.locdata.
        """
        # check if any models are fitted, otherwise it is likely that the fitting procedure was accidentally omitted.
        if self.transformation_models['matrix'] is None and self.transformation_models['offset'] is None:
            raise AttributeError("The transformation_models have to be fitted before they can be evaluated.")

        dimension = locdata.dimension
        frames = locdata.data.frame.values
        matrix = np.tile(np.identity(dimension), (len(frames), *([1] * dimension)))
        offset = np.zeros((len(frames), dimension))

        if self.transformation_models['matrix'] is None:
            transformed_points = locdata.coordinates
        else:
            for n, drift_model in enumerate(self.transformation_models['matrix']):
                matrix[:, n // dimension, n % dimension] = drift_model.eval(frames)
            transformed_points = np.einsum("...ij, ...j -> ...i", matrix, locdata.coordinates)

        if self.transformation_models['offset'] is None:
            pass
        else:
            for n, drift_model in enumerate(self.transformation_models['offset']):
                offset[:, n] = drift_model.eval(frames)
            transformed_points += offset

        return transformed_points

    def apply_correction(self, locdata=None, from_model=True):
        """
        Correct drift by applying the estimated transformations to locdata.

        Parameters
        ----------
        locdata : LocData or None
            Localization data to apply correction on. If None correction is applied to self.locdata.
        from_model : bool
            If `True` compute transformation matrix from fitted transformation models and apply interpolated
            transformations. If False use the estimated transformation matrix for each data chunk.
        """
        if not self:
            logger.warning('No transformations available to be applied.')
            return self

        local_parameter = locals()

        if locdata is None:
            locdata_orig = self.locdata
        else:
            locdata_orig = locdata

        if not len(locdata_orig):
            logger.warning('Locdata is empty.')
            self.locdata_corrected = locdata_orig
            return self

        if from_model:
            transformed_points = self._apply_correction_from_model(locdata=locdata_orig)
        else:
            if locdata is not None:
                raise TypeError('Locdata must be None since correction can only be applied to original locdata chunks.')
            transformed_points = self._apply_correction_on_chunks().coordinates

        # new LocData object
        new_dataframe = locdata_orig.data.copy()
        new_dataframe.update(pd.DataFrame(transformed_points, columns=locdata_orig.coordinate_labels,
                                          index=locdata_orig.data.index))
        new_locdata = LocData.from_dataframe(new_dataframe)

        # update metadata
        meta_ = _modify_meta(self.locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                             parameter=local_parameter,
                             meta=None)
        new_locdata.meta = meta_

        self.locdata_corrected = new_locdata
        return self

    def plot(self, ax=None, transformation_component='matrix', element=None, window=1, **kwargs):
        """
        Plot the transformation components as function of average frame for each locdata chunk.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        transformation_component : str
            One of 'matrix' or 'offset'
        element : int or None
            The element of flattened transformation matrix or offset to be plotted; if None all plots are shown.
        window: int
            Window for running average that is applied before plotting.
            Not implemented yet.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        n_transformations = len(self.transformations)
        # prepare plot
        x = [reference.data.frame.mean() for reference in self.collection.references]
        results = np.array([getattr(transformation, transformation_component) for transformation
                            in self.transformations])
        if element is None:
            ys = results.reshape(n_transformations, -1).T
            for i, y in enumerate(ys):
                ax.plot(x, y, **dict(dict(label=f'{transformation_component}[{i}]'), **kwargs))
        else:
            y = results.reshape(n_transformations, -1).T[element]
            ax.plot(x, y, **dict(dict(label=f'{transformation_component}[{element}]'), **kwargs))

        ax.set(title=f'Drift\n (window={window})',
               xlabel='frame',
               ylabel=''.join([transformation_component])
               )

        return ax



"""

Transform localization data.

This module takes localization data and applies transformation procedures on coordinates or other properties.

"""
import sys

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
import locan.data.hulls
from locan.data.region import Region
from locan.simulation import simulate_uniform
from locan.data.metadata_utils import _modify_meta
from locan.dependencies import HAS_DEPENDENCY

if HAS_DEPENDENCY["open3d"]: import open3d as o3d


__all__ = ['transform_affine', 'randomize']


def _transform_affine_numpy(points, matrix=None, offset=None, pre_translation=None):
    """
    Transform `points` by an affine transformation using standard numpy procedures.

    Parameters
    ----------
    points : array-like
        Points on which to perform the manipulation.
    matrix : tuple with shape (d, d)
        Transformation matrix. If None the unit matrix is used.
    offset : tuple of int or float with shape (d,)
        Translation vector. If None a vector of zeros is used.
    pre_translation : tuple of int or float
        Translation vector for coordinates applied before affine transformation.
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    numpy.ndarray
        Transformed coordinates.
    """
    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else matrix
    offset_ = np.zeros(dimension) if offset is None else offset

    if pre_translation is None:
        # transformed_points = np.array([np.dot(matrix_, point) + offset_ for point in points_])
        # same function but better performance:
        transformed_points = np.einsum("ij, nj -> ni", matrix_, points) + offset_
    else:
        transformed_points = points_ + pre_translation
        # transformed_points = np.array([np.dot(matrix_, point) + offset_ for point in transformed_points])
        transformed_points = np.einsum("ij, nj -> ni", matrix_, transformed_points) + offset_
        transformed_points = transformed_points - pre_translation

    return transformed_points


def _homogeneous_matrix(matrix=np.identity(3), offset=np.zeros(3)):
    """
    Combine transformation matrix and translation vector for dimension d into homogeneous (d+1, d+1) transformation
    matrix to be used with homogeneous coordinates.

    Parameters
    ----------
    matrix : array-like of int or float with shape (d, d)
        Transformation matrix.
    offset : array-like of int or float with shape (d,)
        Translation vector.

    Returns
    -------
    numpy.ndarray with shape (d+1, d+1)
        Homogeneous transformation matrix to be used with homogeneous coordinate vector.
    """
    dimension = np.shape(matrix)[0]

    if dimension != np.shape(matrix)[1]:
        raise ValueError('The matrix has to be of shape (d, d).')

    if dimension != np.shape(offset)[0]:
        raise ValueError('The matrix and offset must have the same dimension.')

    matrix_ = np.identity(dimension+1)
    matrix_[0:dimension, 0:dimension] = matrix
    matrix_[:dimension, dimension] = offset

    return matrix_


def _transform_affine_open3d(points, matrix=None, offset=None, pre_translation=None):
    """
    Transform `points` or coordinates in `locdata` by an affine
    transformation using open3d.

    Parameters
    ----------
    points : array-like
        Points on which to perform the manipulation.
    matrix : tuple with shape (d, d)
        Transformation matrix. If None the unit matrix is used.
    offset : tuple of int or float with shape (d,)
        Translation vector. If None a vector of zeros is used.
    pre_translation : tuple of int or float
        Translation vector for coordinates applied before affine transformation.
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    numpy.ndarray
        Transformed coordinates.
    """
    if not HAS_DEPENDENCY["open3d"]:
        raise ImportError("open3d is required.")

    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else matrix
    offset_ = np.zeros(dimension) if offset is None else offset

    # prepare 3d
    if dimension == 2:
        points_3d = np.concatenate([points_, np.zeros((len(points_), 1))], axis=1)
    elif dimension == 3:
        points_3d = points_
    else:
        raise ValueError('Point array has the wrong shape.')

    # points in open3d
    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points_3d)

    # transform
    matrix_3d = np.identity(3)
    matrix_3d[:dimension, :dimension] = matrix_
    offset_3d = np.zeros(3)
    offset_3d[:dimension] = offset_
    matrix_homogeneous = _homogeneous_matrix(matrix_3d, offset_3d)

    if pre_translation is None:
        point_cloud.transform(matrix_homogeneous)
    else:
        pre_translation_3d = np.zeros(3)
        pre_translation_3d[:dimension] = pre_translation
        point_cloud.translate(pre_translation_3d)
        point_cloud.transform(matrix_homogeneous)
        point_cloud.translate(-pre_translation_3d)

    # return correct dimension
    if dimension == 2:
        transformed_points = np.asarray(point_cloud.points)[:, 0:2]
    else:  # if dimension == 3:
        transformed_points = np.asarray(point_cloud.points)

    return transformed_points


def transform_affine(locdata, matrix=None, offset=None, pre_translation=None, method='numpy'):
    """
    Transform `points` or coordinates in `locdata` by an affine transformation.

    Parameters
    ----------
    locdata : numpy.ndarray or LocData
        Localization data on which to perform the manipulation.
    matrix :
        Transformation matrix. If None the unit matrix is used.
    offset : tuple of int or float
        Values for translation. If None a vector of zeros is used.
    pre_translation : tuple of int or float
        Translation vector for coordinates applied before affine transformation.
        The reverse translation is applied after the affine transformation.
    method : string
        The method (i.e. library or algorithm) used for computation. One of 'numpy', 'open3d'.

    Returns
    -------
    numpy.ndarray or LocData
        New localization data with transformed coordinates.
    """
    local_parameter = locals()

    if len(locdata) == 0:
        return locdata

    # adjust input
    if isinstance(locdata, LocData):
        points = locdata.coordinates
    else:
        points = locdata

    if method == 'numpy':
        transformed_points = _transform_affine_numpy(points, matrix=matrix, offset=offset,
                                                     pre_translation=pre_translation)
    elif method == 'open3d':
        transformed_points = _transform_affine_open3d(points, matrix=matrix, offset=offset,
                                                      pre_translation=pre_translation)
    else:
        raise ValueError(f'Method {method} is not available.')

    # prepare output
    if isinstance(locdata, LocData):
        # new LocData object
        new_dataframe = locdata.data.copy()
        new_dataframe.update(pd.DataFrame(transformed_points, columns=locdata.coordinate_labels,
                                          index=locdata.data.index))
        new_locdata = LocData.from_dataframe(new_dataframe)

        # update metadata
        meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                             parameter=local_parameter,
                             meta=None)
        new_locdata.meta = meta_

        return new_locdata

    else:
        return transformed_points


def randomize(locdata, hull_region='bb', seed=None):
    """
    Transform locdata coordinates into randomized coordinates that follow complete spatial randomness on the same
    region as the input locdata.

    Parameters
    ----------
    locdata : LocData
        Localization data to be randomized
    hull_region : Region, str
        Region of interest. String identifier can be one of 'bb', 'ch', 'as', 'obb' referring to the corresponding hull.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    locdata : LocData
        New localization data with randomized coordinates.
    """
    local_parameter = locals()

    rng = np.random.default_rng(seed)

    if hull_region == 'bb':
        region_ = locdata.bounding_box.hull.T
    elif hull_region == 'ch':
        region_ = locdata.convex_hull.region
    elif hull_region == 'as':
        region_ = locdata.alpha_shape.region
    elif hull_region == 'obb':
        region_ = locdata.oriented_bounding_box.region
    elif isinstance(hull_region, Region):
        region_ = hull_region
    else:
        raise NotImplementedError

    new_locdata = simulate_uniform(n_samples=len(locdata), region=region_, seed=rng)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter, meta=None)
    new_locdata.meta = meta_

    return new_locdata



import pytest
import numpy as np
import matplotlib.pyplot as plt  # needed for visualization

from locan import LocData
from locan.analysis import RipleysKFunction, RipleysLFunction, RipleysHFunction


def test_Ripleys_k_function_empty(caplog):
    rhf = RipleysKFunction().compute(LocData())
    rhf.plot()
    assert caplog.record_tuples == [('locan.analysis.ripley', 30, 'Locdata is empty.')]


def test_Ripleys_k_function(locdata_2d):
    radii = np.linspace(0, 20, 11)
    rhf = RipleysKFunction(radii=radii).compute(locdata_2d)
    # print(rhf.results)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    rhf.plot()
    # plt.show()

    other_locdata = LocData.from_selection(locdata_2d, indices=[0, 1, 2])
    rhf = RipleysKFunction(radii=radii, region_measure=1).compute(locdata_2d, other_locdata=other_locdata)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    plt.close('all')


def test_Ripleys_k_function_3d(locdata_3d):
    radii = np.linspace(0, 20, 20)
    rhf = RipleysKFunction(radii=radii).compute(locdata_3d)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    rhf.plot()
    # plt.show()

    other_locdata = LocData.from_selection(locdata_3d, indices=[0, 1, 2])
    rhf = RipleysKFunction(radii=radii, region_measure=1).compute(locdata_3d, other_locdata=other_locdata)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    plt.close('all')


def test_Ripleys_l_function(locdata_2d):
    radii = np.linspace(0, 20, 20)
    rhf = RipleysLFunction(radii=radii).compute(locdata_2d)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    rhf.plot()
    # plt.show()

    other_locdata = LocData.from_selection(locdata_2d, indices=[0, 1, 2])
    rhf = RipleysLFunction(radii=radii, region_measure=1).compute(locdata_2d, other_locdata=other_locdata)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    plt.close('all')


def test_Ripleys_h_function(locdata_2d):
    radii = np.linspace(0, 20, 21)
    rhf = RipleysHFunction(radii=radii).compute(locdata_2d)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    rhf.plot()
    # plt.show()

    other_locdata = LocData.from_selection(locdata_2d, indices=[0, 1, 2])
    rhf = RipleysHFunction(radii=radii, region_measure=1).compute(locdata_2d, other_locdata=other_locdata)
    assert all(rhf.results.index == radii)
    assert (len(rhf.results) == len(radii))

    assert len(rhf.Ripley_h_maximum) == 1
    assert rhf.Ripley_h_maximum.iloc[0].radius == 0
    assert rhf.Ripley_h_maximum.iloc[0].Ripley_h_maximum == 0
    del rhf.Ripley_h_maximum
    assert rhf.Ripley_h_maximum.iloc[0].radius == 0
    assert rhf.Ripley_h_maximum.iloc[0].Ripley_h_maximum == 0

    plt.close('all')



"""

User interfaces.

This module provides functions and classes for using graphical user interfaces (GUI).
Functions provide a GUI based on QT if the QT backend and appropriate python bindings are available.

The constant `locan.QT_BINDINGS` declares which python binding to use.
Depending on the situation it is set to one of the values specified in the enum class `locan.QtBindings`.

1) `QtBindings.NONE` if no QT backend with python bindings is installed.

2) `QtBindings.PYSIDE2` or `QtBindings.PYQT5` if exactly one of them is installed.

3) the `QtBindings` value that is equal to the python environment variable `QT_API`
   if both PySide2 and PyQt5 are installed and `QT_API` exists.

4) `QtBindings.PYSIDE2` if both PySide2 and PyQt5 are installed and `QT_API` does not exist.

To force locan to use specific QT bindings set the `QT_API` environment variable to 'pyside2' or 'pyqt5'.


Submodules:
-----------

.. autosummary::
   :toctree: ./

   io

"""
from .io import *

__all__ = []
__all__.extend(io.__all__)


"""

Compute localization precision from successive nearby localizations.

Localization precision is estimated from spatial variations of all localizations that appear in successive frames
within a specified search radius [1]_.

Localization pair distance distributions are fitted according to the probability density functions in [2]_.

The estimated sigmas describe the standard deviation for pair distances. Localization precision is often defined as
the standard deviation for localization distances from the center position. With that definition, the localization
precision is equal to sigma / sqrt(2).

References
----------
.. [1] Endesfelder, Ulrike, et al., A simple method to estimate the average localization precision of a single-molecule
   localization microscopy experiment. Histochemistry and Cell Biology 141.6 (2014): 629-638.

.. [2] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
   A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
   Biophysical Journal 90 (2), 2006, 668-671,
   doi.org/10.1529/biophysj.105.065599.

"""

import warnings
import logging

import numpy as np
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt

from sklearn.neighbors import NearestNeighbors
from scipy import stats

from locan.configuration import N_JOBS, TQDM_DISABLE, TQDM_LEAVE
from locan.analysis.analysis_base import _Analysis, _list_parameters


__all__ = ['LocalizationPrecision']

logger = logging.getLogger(__name__)


##### The algorithms

def _localization_precision(locdata, radius=50):
    # group localizations
    grouped = locdata.data.groupby('frame')

    # find nearest neighbors
    min = locdata.data['frame'].unique().min()
    max = locdata.data['frame'].unique().max()

    results = pd.DataFrame()

    for i in tqdm(range(min, max - 1),
                  desc='Processed frames:', leave=TQDM_LEAVE, disable=TQDM_DISABLE):
        try:
            points = grouped.get_group(i)[locdata.coordinate_labels]
            other_points = grouped.get_group(i + 1)[locdata.coordinate_labels]

            # print(points)

            nn = NearestNeighbors(radius=radius, metric='euclidean', n_jobs=N_JOBS).fit(other_points)
            distances, indices = nn.radius_neighbors(points)

            if len(distances):
                for n, (dists, inds) in enumerate(zip(distances, indices)):
                    if len(dists):
                        min_distance = np.amin(dists)
                        min_position = np.argmin(dists)
                        min_index = inds[min_position]
                        difference = points.iloc[n] - other_points.iloc[min_index]

                        df = difference.to_frame().T
                        df = df.rename(columns={'position_x': 'position_delta_x',
                                                'position_y': 'position_delta_y',
                                                'position_z': 'position_delta_z'})
                        df = df.assign(position_distance=min_distance)
                        df = df.assign(frame=i)
                        results = pd.concat([results, df])
        except KeyError:
            pass

    results.reset_index(inplace=True, drop=True)
    return results


##### The specific analysis classes


class LocalizationPrecision(_Analysis):
    """
    Compute the localization precision from consecutive nearby localizations.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radius : int, float
        Search radius for nearest-neighbor searches.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : numpy.ndarray, pandas.DataFrame
        Computed results.
    distribution_statistics : Distribution_fits object, None
        Distribution parameters derived from MLE fitting of results.
    """
    def __init__(self, meta=None, radius=50):
        super().__init__(meta=meta, radius=radius)
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _localization_precision(locdata=locdata, **self.parameter)
        if self.results.empty:
            logger.warning('No successive localizations were found.')

        return self

    def fit_distributions(self, loc_property=None, **kwargs):
        """
        Fit probability density functions to the distributions of `loc_property` values in the results
        using MLE (scipy.stats).

        Parameters
        ----------
        loc_property : str
            The LocData property for which to fit an appropriate distribution; if None all plots are shown.
        kwargs : dict
            Other parameters passed to the `distribution.fit()` method.
        """
        if not self:
            logger.warning('No results available to be fitted.')
            return

        self.distribution_statistics = _DistributionFits(self)

        if loc_property is None:
            for prop in ['position_delta_x', 'position_delta_y', 'position_delta_z', 'position_distance']:
                if prop in self.results.columns:
                    self.distribution_statistics.fit(loc_property=prop, **kwargs)
        else:
            self.distribution_statistics.fit(loc_property=loc_property, **kwargs)

    def plot(self, ax=None, loc_property=None, window=1, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the running average of results over window size.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        loc_property : str, list(str)
            The property for which to plot localization precision; if None all plots are shown.
        window: int
            Window for running average that is applied before plotting.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        # prepare plot
        self.results.rolling(window=window, center=True).mean().plot(ax=ax,
                                                                     x='frame',
                                                                     y=loc_property,
                                                                     **dict(dict(legend=False), **kwargs))
        ax.set(title=f'Localization Precision\n (window={window})',
               xlabel='frame',
               ylabel=loc_property
               )

        return ax


    def hist(self, ax=None, loc_property='position_distance', bins='auto', fit=True, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing the distributions of results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        loc_property : str
            The property for which to plot localization precision.
        bins : float
            Bin specifications (passed to :func:`matplotlib.hist`).
        fit: Bool
            Flag indicating if distributions fit are shown.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.his`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        # prepare plot
        ax.hist(self.results[loc_property].values, bins=bins, **dict(dict(density=True, log=False), **kwargs))
        ax.set(title='Localization Precision',
               xlabel=loc_property,
               ylabel='PDF'
               )

        if fit:
            if isinstance(self.distribution_statistics, _DistributionFits):
                self.distribution_statistics.plot(ax=ax, loc_property=loc_property)
            else:
                self.fit_distributions()
                self.distribution_statistics.plot(ax=ax, loc_property=loc_property)

        return ax


#### Auxiliary functions and classes

class PairwiseDistance1d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 1D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    mu : float
        Distance between the point cloud center positions.
    sigma_1 : float
        Standard deviation for one point cloud.
    sigma_2 : float
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, mu, sigma_1, sigma_2):
        sigma = np.sqrt(sigma_1 ** 2 + sigma_2 ** 2)
        return np.sqrt(2 / np.pi) / sigma * np.exp(- (mu ** 2 + x ** 2) / (2 * sigma ** 2)) \
               * np.cosh(x * mu / (sigma ** 2))


class PairwiseDistance2d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    mu : float
        Distance between the point cloud center positions.
    sigma_1 : float
        Standard deviation for one point cloud.
    sigma_2 : float
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, mu, sigma_1, sigma_2):
        sigma = np.sqrt(sigma_1 ** 2 + sigma_2 ** 2)
        return x / (sigma ** 2) * np.exp(- (mu ** 2 + x ** 2) / (2 * sigma ** 2)) * np.i0(x * mu / (sigma ** 2))


class PairwiseDistance2dIdenticalSigma(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    mu : float
        Distance between the point cloud center positions.
    sigma : float
        Standard deviation for both point clouds.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, mu, sigma):
        return x / (sigma ** 2) * np.exp(- (mu ** 2 + x ** 2) / (2 * sigma ** 2)) * np.i0(x * mu / (sigma ** 2))


class PairwiseDistance3d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 3D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    mu : float
        Distance between the point cloud center positions.
    sigma_1 : float
        Standard deviation for one point cloud.
    sigma_2 : float
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, mu, sigma_1, sigma_2):
        sigma = np.sqrt(sigma_1 ** 2 + sigma_2 ** 2)
        if all(mu == 0):
            return np.sqrt(2 / np.pi) * x / sigma * np.exp(- (mu ** 2 + x ** 2) / (2 * sigma ** 2)) * x / (sigma ** 2)
        else:
            return np.sqrt(2 / np.pi) * x / sigma / mu * np.exp(- (mu ** 2 + x ** 2) / (2 * sigma ** 2)) * np.sinh(
                x * mu / (sigma ** 2))


class PairwiseDistance1dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 1D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    sigma : float
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, sigma):
        return np.sqrt(2 / np.pi) / sigma * np.exp(- x ** 2 / (2 * sigma ** 2))


class PairwiseDistance2dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    sigma : float
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, sigma):
        return x / (sigma ** 2) * np.exp(- x ** 2 / (2 * sigma ** 2))


class PairwiseDistance3dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 3D (as described in [1]_).

    The continuous distribution class inherits from scipy.stats.rv_continuous a set of methods and is defined by
    overriding the _pdf method.

    Parameters
    ----------
    x : float
        distance
    sigma : float
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """
    def _pdf(self, x, sigma):
        return np.sqrt(2 / np.pi) * x / sigma * np.exp(- x ** 2 / (2 * sigma ** 2)) * x / (sigma ** 2)


class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by LocalizationPrecision methods.
    It holds the statistical parameters derived by fitting the result distributions using MLE (scipy.stats).

    Parameters
    ----------
    analyis_class : LocalizationPrecision object
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationPrecision object
        The analysis class with result data to fit.
    pairwise_distribution : Pairwise_distance_distribution_2d
        Continuous distribution function used to fit Position_distances
    parameters : list of str
        Distribution parameters.

    Note
    ----
    Attributes for fit parameter are generated dynamically, named as loc_property + distribution parameters and listed
    in parameters.
    """

    def __init__(self, analysis_class):
        self.analysis_class = analysis_class
        self.distribution = None
        self._dist_parameters = None
        self.parameters = []

        # continuous distributions
        if self.analysis_class.results is None:
            self.pairwise_distribution = None
        else:
            delta_columns = [c for c in self.analysis_class.results.columns if 'position_delta' in c]
            if len(delta_columns) == 1:
                self.pairwise_distribution = PairwiseDistance1dIdenticalSigmaZeroMu(name='pairwise', a=0.)
            elif len(delta_columns) == 2:
                self.pairwise_distribution = PairwiseDistance2dIdenticalSigmaZeroMu(name='pairwise', a=0.)
            elif len(delta_columns) == 3:
                self.pairwise_distribution = PairwiseDistance3dIdenticalSigmaZeroMu(name='pairwise', a=0.)
            # a is the lower bound of the support of the distribution
            # self.pairwise_distribution = PairwiseDistance2dIdenticalSigma(name='pairwise') also works but is very slow.

    def fit(self, loc_property='position_distance', **kwargs):
        """
        Fit distributions of results using a MLE fit (scipy.stats) and provide fit results.

        Parameters
        ----------
        loc_property : str
            The property for which to fit an appropriate distribution
        kwargs : dict
            Other parameters passed to the `distribution.fit()` method.
        """
        # prepare parameters
        if 'position_delta_' in loc_property:
            self.distribution = stats.norm
            self._dist_parameters = [(loc_property + '_' + param) for param in ['loc', 'scale']]
        elif loc_property == 'position_distance':
            self.distribution = self.pairwise_distribution
            self._dist_parameters = [(loc_property + '_' + param) for param in ['sigma', 'loc', 'scale']]
        else:
            raise TypeError('Unknown localization property.')

        for param in self._dist_parameters:
            if param not in self.parameters:
                self.parameters.append(param)

        # MLE fit of distribution on data
        if 'position_delta_' in loc_property:
            fit_results = self.distribution.fit(self.analysis_class.results[loc_property].values, **kwargs)
        elif loc_property == 'position_distance':
            fit_results = self.distribution.fit(self.analysis_class.results[loc_property].values,
                                                **dict(dict(floc=0, fscale=1), **kwargs))
        else:
            raise TypeError('Unknown localization property.')

        for parameter, result in zip(self._dist_parameters, fit_results):
            setattr(self, parameter, result)

    def plot(self, ax=None, loc_property='position_distance', **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the probability distribution functions of fitted
        results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        loc_property : str
            The property for which to plot the distribution fit.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        if 'position_delta_' in loc_property:
            _loc = getattr(self, loc_property + '_loc')
            _scale = getattr(self, loc_property + '_scale')

            x_values = np.linspace(stats.norm.ppf(0.01, loc=_loc, scale=_scale),
                                   stats.norm.ppf(0.99, loc=_loc, scale=_scale), 100)
            ax.plot(x_values, stats.norm.pdf(x_values, loc=_loc, scale=_scale), 'r-', lw=3, alpha=0.6,
                    label='fitted pdf', **kwargs)

        elif loc_property == 'position_distance':
            if isinstance(self.pairwise_distribution, PairwiseDistance2dIdenticalSigma):  # pragma: no cover
                _sigma = self.position_distance_sigma
                _mu = self.position_distance_mu
                x_values = np.linspace(self.pairwise_distribution.ppf(0.01, mu=_mu, sigma=_sigma),
                                       self.pairwise_distribution.ppf(0.99, mu=_mu, sigma=_sigma), 100)
                ax.plot(x_values, self.pairwise_distribution.pdf(x_values, mu=_mu, sigma=_sigma), 'r-', lw=3, alpha=0.6,
                        label='fitted pdf', **kwargs)
            elif isinstance(self.pairwise_distribution, (PairwiseDistance1dIdenticalSigmaZeroMu,
                                                         PairwiseDistance2dIdenticalSigmaZeroMu,
                                                         PairwiseDistance3dIdenticalSigmaZeroMu)):
                _sigma = self.position_distance_sigma
                x_values = np.linspace(self.pairwise_distribution.ppf(0.01, sigma=_sigma),
                                       self.pairwise_distribution.ppf(0.99, sigma=_sigma), 100)
                ax.plot(x_values, self.pairwise_distribution.pdf(x_values, sigma=_sigma), 'r-', lw=3, alpha=0.6,
                        label='fitted pdf', **kwargs)
            else:
                raise NotImplementedError('pairwise_distribution function has not been implemented for plotting '
                                          'position distances.')

        return ax

    def parameter_dict(self):
        """ Dictionary of fitted parameters. """
        return {k: self.__dict__[k] for k in self.parameters}



from io import StringIO

import locan.constants
from locan.locan_io import load_Elyra_file
from locan.locan_io.locdata.elyra_io import load_Elyra_header


def test_get_correct_column_names_from_Elyra_header():
    columns = load_Elyra_header(path=locan.ROOT_DIR / 'tests/test_data/Elyra_dstorm_data.txt')
    assert (columns == ['original_index', 'frame', 'frames_number', 'frames_missing', 'position_x', 'position_y',
                        'uncertainty', 'intensity', 'local_background_sigma', 'chi_square', 'psf_half_width', 'channel',
                        'slice_z'])

    file_like = StringIO("Index	First Frame	Number Frames	Frames Missing	Position X [nm]	Position Y [nm]\n"
                         "1  1   1   0   15850.6 23502.1")
    columns = load_Elyra_header(path=file_like)
    assert (columns == ['original_index', 'frame', 'frames_number', 'frames_missing', 'position_x', 'position_y'])


def test_loading_Elyra_file():
    dat = load_Elyra_file(path=locan.ROOT_DIR / 'tests/test_data/Elyra_dstorm_data.txt')
    # loading is not limited by nrows=10 to ensure correct treatment of file appendix and NUL character.
    assert (len(dat) == 999)

    file_like = StringIO("Index\tFirst Frame\tNumber Frames\tFrames Missing\tPosition X [nm]\tPosition Y [nm]\n"
                         "1\t1\t1\t0\t15850.6\t23502.1")
    dat = load_Elyra_file(path=file_like)
    # loading is not limited by nrows=10 to ensure correct treatment of file appendix and NUL character.
    assert (len(dat) == 1)



from io import StringIO

import numpy as np
import locan.constants
from locan.locan_io import load_rapidSTORM_file, load_rapidSTORM_track_file

from locan.locan_io.locdata.rapidstorm_io import load_rapidSTORM_header, load_rapidSTORM_track_header


def test_get_correct_column_names_from_rapidSTORM_header():
    columns = load_rapidSTORM_header(path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt')
    assert columns == ['position_x', 'position_y', 'frame', 'intensity', 'chi_square', 'local_background']

    file_like = StringIO('# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations>\n'
                         '9657.4 24533.5 0 33290.1 1.19225e+006 767.733')
    columns = load_rapidSTORM_header(path=file_like)
    assert columns == ['position_x', 'position_y', 'frame', 'intensity', 'chi_square', 'local_background']


def test_loading_rapidSTORM_file():
    dat = load_rapidSTORM_file(path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt',
                                  nrows=10)
    # print(dat.data.head())
    # dat.print_meta()
    assert (len(dat) == 10)

    file_like = StringIO('# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations>\n'
                         '9657.4 24533.5 0 33290.1 1.19225e+006 767.733')
    dat = load_rapidSTORM_file(path=file_like, nrows=1)
    assert (len(dat) == 1)


def test_get_correct_column_names_from_rapidSTORM_track_header():
    columns = load_rapidSTORM_track_header(
        path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_track_data.txt')
    assert columns == (
        ['position_x', 'position_y', 'frame', 'intensity'],
        ['position_x', 'uncertainty_x', 'position_y', 'uncertainty_y',
         'frame', 'intensity', 'chi_square', 'local_background']
    )

    file_like = StringIO('# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-0-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in X" unit="nanometer" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in Y" unit="nanometer" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations></localizations>\n'
                         '5417.67 8439.85 1 4339.29 2 5421.22 25 8440.4 25 0 2292.39 48696.1 149.967 5414.13 25 8439.3 25 1 2046.9 65491.4 142.521'
                         )
    columns = load_rapidSTORM_track_header(path=file_like)
    assert columns == (
        ['position_x', 'position_y', 'frame', 'intensity'],
        ['position_x', 'uncertainty_x', 'position_y', 'uncertainty_y',
         'frame', 'intensity', 'chi_square', 'local_background']
    )


def test_loading_rapidSTORM_track_file():
    dat = load_rapidSTORM_track_file(
        path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_track_data.txt',
        min_localization_count=2,
        nrows=10)
    # print(dat.data.head())
    # print(dat.data.columns)
    # dat.print_meta()
    assert np.array_equal(dat.data.columns,
                          ['localization_count', 'position_x', 'position_y',
                          'region_measure_bb', 'localization_density_bb', 'subregion_measure_bb']
                          )
    assert (len(dat) == 9)  # len(dat) is 9 and not 10 since one row is filtered out y min_localization_count=2

    dat = load_rapidSTORM_track_file(
        path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_track_data.txt',
        collection=False,
        nrows=10)
    assert np.array_equal(dat.data.columns,
                          ['position_x', 'position_y', 'frame', 'intensity']
                          )
    assert (len(dat) == 10)

    file_like = StringIO('# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-0-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in X" unit="nanometer" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in Y" unit="nanometer" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations></localizations>\n'
                         '5417.67 8439.85 1 4339.29 2 5421.22 25 8440.4 25 0 2292.39 48696.1 149.967 5414.13 25 8439.3 25 1 2046.9 65491.4 142.521'
                         )
    dat = load_rapidSTORM_file(path=file_like, nrows=1)
    assert (len(dat) == 1)



"""
Compute additional properties for locdata objects.

These functions take locdata as input, and compute one or more new properties.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   misc
   statistics
"""

from locan.data.properties.misc import *
from locan.data.properties.locdata_statistics import *

__all__ = []
__all__.extend(misc.__all__)
__all__.extend(locdata_statistics.__all__)


"""

Transform localization data.

This module provides functions that take locdata as input, transform the localization data,
and return a new LocData object.


Submodules:
-----------

.. autosummary::
   :toctree: ./

   bunwarpj
   transformation

"""

from locan.data.transform.bunwarpj import *
from locan.data.transform.transformation import *

__all__ = []
__all__.extend(bunwarpj.__all__)
__all__.extend(transformation.__all__)


import numpy as np
import pytest

import locan as lc
from locan.dependencies import HAS_DEPENDENCY
from locan import load_decode_header, load_decode_file


pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")


def test_load_DECODE_header():
    columns, meta, decode = load_decode_header(path=lc.ROOT_DIR / 'tests/test_data/decode_dstorm_data.h5')
    # print(columns, meta, decode)
    assert list(meta.keys()) == ['px_size', 'xy_unit']
    assert list(decode.keys()) == ['version']
    assert columns == [
        'local_background',
        'bg_cr',
        'bg_sig',
        'frame',
        'original_index',
        'intensity',
        'phot_cr',
        'phot_sig',
        'prob',
        'position_x',
        'position_y',
        'position_z',
        'x_cr',
        'y_cr',
        'z_cr',
        'x_sig',
        'y_sig',
        'z_sig'
    ]


def test_loading_DECODE_file_empty_file():
    locdata = load_decode_file(
        path=lc.ROOT_DIR / 'tests/test_data/decode_dstorm_data_empty.h5',
        nrows=10)
    assert len(locdata) == 0


def test_loading_DECODE_file():
    locdata = load_decode_file(
        path=lc.ROOT_DIR / 'tests/test_data/decode_dstorm_data.h5',
        nrows=10)
    # print(locdata.data)
    assert len(locdata) == 10
    assert np.array_equal(locdata.data.columns, [
        'local_background', 'frame', 'original_index', 'intensity', 'phot_sig',
        'prob', 'position_x', 'position_y', 'position_z', 'x_sig', 'y_sig',
        'z_sig'
    ])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: locan/data/metadata.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19locan/data/metadata.proto\x12\nlocan.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\",\n\tOperation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tparameter\x18\x02 \x01(\t\"n\n\x04\x46ile\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12#\n\x04type\x18\x03 \x01(\x0e\x32\x15.locan.data.File_type\x12\x0c\n\x04path\x18\x04 \x01(\t\x12\x0e\n\x06groups\x18\x05 \x03(\t\"R\n\x07\x41\x64\x64ress\x12\x15\n\raddress_lines\x18\x01 \x03(\t\x12\x0c\n\x04\x63ity\x18\x02 \x01(\t\x12\x11\n\tcity_code\x18\x03 \x01(\t\x12\x0f\n\x07\x63ountry\x18\x04 \x01(\t\"Z\n\x0b\x41\x66\x66iliation\x12\x11\n\tinstitute\x18\x01 \x01(\t\x12\x12\n\ndepartment\x18\x02 \x01(\t\x12$\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x13.locan.data.Address\"\xd7\x01\n\x06Person\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x12\n\nfirst_name\x18\x03 \x01(\t\x12\x11\n\tlast_name\x18\x04 \x01(\t\x12\r\n\x05title\x18\x05 \x01(\t\x12-\n\x0c\x61\x66\x66iliations\x18\x06 \x03(\x0b\x32\x17.locan.data.Affiliation\x12$\n\x07\x61\x64\x64ress\x18\x07 \x01(\x0b\x32\x13.locan.data.Address\x12\x0e\n\x06\x65mails\x18\x08 \x03(\t\x12\r\n\x05roles\x18\t \x03(\t\"\xd3\x01\n\x12\x45xperimentalSample\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x0f\n\x07targets\x18\x03 \x03(\t\x12\x14\n\x0c\x66luorophores\x18\x04 \x03(\t\x12\x0f\n\x07\x62uffers\x18\x05 \x03(\t\x12\x34\n\x03map\x18\x06 \x03(\x0b\x32\'.locan.data.ExperimentalSample.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc9\x01\n\x11\x45xperimentalSetup\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12.\n\roptical_units\x18\x03 \x03(\x0b\x32\x17.locan.data.OpticalUnit\x12\x33\n\x03map\x18\x04 \x03(\x0b\x32&.locan.data.ExperimentalSetup.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe6\x01\n\x0bOpticalUnit\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12.\n\x0cillumination\x18\x03 \x01(\x0b\x32\x18.locan.data.Illumination\x12(\n\tdetection\x18\x04 \x01(\x0b\x32\x15.locan.data.Detection\x12,\n\x0b\x61\x63quisition\x18\x05 \x01(\x0b\x32\x17.locan.data.Acquisition\x12*\n\nlightsheet\x18\x06 \x01(\x0b\x32\x16.locan.data.Lightsheet\"\xec\x01\n\x0cIllumination\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x13\n\x0blightsource\x18\x03 \x01(\t\x12\r\n\x05power\x18\x04 \x01(\x02\x12\x0c\n\x04\x61rea\x18\x05 \x01(\x02\x12\x15\n\rpower_density\x18\x06 \x01(\x02\x12\x12\n\nwavelength\x18\x07 \x01(\x02\x12.\n\x03map\x18\x08 \x03(\x0b\x32!.locan.data.Illumination.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xad\x01\n\tDetection\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\"\n\x06\x63\x61mera\x18\x03 \x01(\x0b\x32\x12.locan.data.Camera\x12+\n\x03map\x18\x04 \x03(\x0b\x32\x1e.locan.data.Detection.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x92\x03\n\x06\x43\x61mera\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\r\n\x05model\x18\x04 \x01(\t\x12\x0c\n\x04gain\x18\x05 \x01(\x02\x12\x1b\n\x13\x65lectrons_per_count\x18\x06 \x01(\x02\x12\x33\n\x10integration_time\x18\x07 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x15\n\rpixel_count_x\x18\x08 \x01(\x05\x12\x15\n\rpixel_count_y\x18\t \x01(\x05\x12\x14\n\x0cpixel_size_x\x18\n \x01(\x02\x12\x14\n\x0cpixel_size_y\x18\x0b \x01(\x02\x12\x0f\n\x07\x66lipped\x18\x0c \x01(\x08\x12(\n\x03map\x18\r \x03(\x0b\x32\x1b.locan.data.Camera.MapEntry\x12\x0e\n\x06offset\x18\x0e \x01(\x02\x12\x15\n\rserial_number\x18\x0f \x01(\t\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x89\x03\n\x0b\x41\x63quisition\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x13\n\x0b\x66rame_count\x18\x03 \x01(\x05\x12\x1f\n\x17\x66rame_of_interest_first\x18\x04 \x01(\x05\x12\x1e\n\x16\x66rame_of_interest_last\x18\x05 \x01(\x05\x12.\n\ntime_start\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08time_end\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0bstack_count\x18\x08 \x01(\x05\x12\x18\n\x10stack_step_count\x18\t \x01(\x05\x12\x17\n\x0fstack_step_size\x18\n \x01(\x02\x12-\n\x03map\x18\x0b \x03(\x0b\x32 .locan.data.Acquisition.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xbe\x01\n\nLightsheet\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x0f\n\x07\x61ngle_x\x18\x03 \x01(\x02\x12\x0f\n\x07\x61ngle_y\x18\x04 \x01(\x02\x12\x0f\n\x07\x61ngle_z\x18\x05 \x01(\x02\x12,\n\x03map\x18\x06 \x03(\x0b\x32\x1f.locan.data.Lightsheet.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x96\x02\n\nExperiment\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12)\n\rexperimenters\x18\x03 \x03(\x0b\x32\x12.locan.data.Person\x12/\n\x07samples\x18\x04 \x03(\x0b\x32\x1e.locan.data.ExperimentalSample\x12-\n\x06setups\x18\x05 \x03(\x0b\x32\x1d.locan.data.ExperimentalSetup\x12,\n\x03map\x18\x06 \x03(\x0b\x32\x1f.locan.data.Experiment.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xdd\x01\n\tLocalizer\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x10\n\x08software\x18\x03 \x01(\t\x12\x1b\n\x13intensity_threshold\x18\x04 \x01(\x02\x12\x11\n\tpsf_fixed\x18\x05 \x01(\x08\x12\x10\n\x08psf_size\x18\x06 \x01(\x02\x12+\n\x03map\x18\x07 \x03(\x0b\x32\x1e.locan.data.Localizer.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa7\x01\n\x08Relation\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x1e\n\x04\x66ile\x18\x03 \x01(\x0b\x32\x10.locan.data.File\x12*\n\x03map\x18\x04 \x03(\x0b\x32\x1d.locan.data.Relation.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xb1\x01\n\x08Property\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0c\n\x04unit\x18\x04 \x01(\t\x12\x0c\n\x04type\x18\x05 \x01(\t\x12*\n\x03map\x18\x06 \x03(\x0b\x32\x1d.locan.data.Property.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xf5\x06\n\x08Metadata\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\"\n\x06source\x18\x06 \x01(\x0e\x32\x12.locan.data.Source\x12 \n\x05state\x18\x07 \x01(\x0e\x32\x11.locan.data.State\x12&\n\x07history\x18\x08 \x03(\x0b\x32\x15.locan.data.Operation\x12\x1c\n\x14\x61ncestor_identifiers\x18\t \x03(\t\x12(\n\nproperties\x18\x12 \x03(\x0b\x32\x14.locan.data.Property\x12\x35\n\x17localization_properties\x18\x18 \x03(\x0b\x32\x14.locan.data.Property\x12\x15\n\relement_count\x18\x0b \x01(\x03\x12\x13\n\x0b\x66rame_count\x18\x0c \x01(\x03\x12\x1e\n\x04\x66ile\x18\x14 \x01(\x0b\x32\x10.locan.data.File\x12\'\n\trelations\x18\x15 \x03(\x0b\x32\x14.locan.data.Relation\x12*\n\nexperiment\x18\x16 \x01(\x0b\x32\x16.locan.data.Experiment\x12(\n\tlocalizer\x18\x17 \x01(\x0b\x32\x15.locan.data.Localizer\x12*\n\x03map\x18\x11 \x03(\x0b\x32\x1d.locan.data.Metadata.MapEntry\x12\x31\n\rcreation_time\x18\x19 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x11modification_time\x18\x1a \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x33\n\x0fproduction_time\x18\x1b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x03\x10\x06J\x04\x08\n\x10\x0bJ\x04\x08\r\x10\x11J\x04\x08\x13\x10\x14R\x04unitR\tfile_typeR\tfile_pathR\x12\x65xperimental_setupR\x13\x65xperimental_sampleR\x05unitsR\rcreation_dateR\x11modification_dateR\x0fproduction_date*T\n\x06Source\x12\x12\n\x0eUNKNOWN_SOURCE\x10\x00\x12\n\n\x06\x44\x45SIGN\x10\x01\x12\x0e\n\nEXPERIMENT\x10\x02\x12\x0e\n\nSIMULATION\x10\x03\x12\n\n\x06IMPORT\x10\x04*1\n\x05State\x12\x11\n\rUNKNOWN_STATE\x10\x00\x12\x07\n\x03RAW\x10\x01\x12\x0c\n\x08MODIFIED\x10\x02*\xaa\x01\n\tFile_type\x12\x15\n\x11UNKNOWN_FILE_TYPE\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0e\n\nRAPIDSTORM\x10\x02\x12\t\n\x05\x45LYRA\x10\x03\x12\x10\n\x0cTHUNDERSTORM\x10\x04\x12\x08\n\x04\x41SDF\x10\x05\x12\x0e\n\nNANOIMAGER\x10\x06\x12\x13\n\x0fRAPIDSTORMTRACK\x10\x07\x12\x08\n\x04SMLM\x10\x08\x12\n\n\x06\x44\x45\x43ODE\x10\t\x12\x08\n\x04SMAP\x10\n')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'locan.data.metadata_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXPERIMENTALSAMPLE_MAPENTRY._options = None
  _EXPERIMENTALSAMPLE_MAPENTRY._serialized_options = b'8\001'
  _EXPERIMENTALSETUP_MAPENTRY._options = None
  _EXPERIMENTALSETUP_MAPENTRY._serialized_options = b'8\001'
  _ILLUMINATION_MAPENTRY._options = None
  _ILLUMINATION_MAPENTRY._serialized_options = b'8\001'
  _DETECTION_MAPENTRY._options = None
  _DETECTION_MAPENTRY._serialized_options = b'8\001'
  _CAMERA_MAPENTRY._options = None
  _CAMERA_MAPENTRY._serialized_options = b'8\001'
  _ACQUISITION_MAPENTRY._options = None
  _ACQUISITION_MAPENTRY._serialized_options = b'8\001'
  _LIGHTSHEET_MAPENTRY._options = None
  _LIGHTSHEET_MAPENTRY._serialized_options = b'8\001'
  _EXPERIMENT_MAPENTRY._options = None
  _EXPERIMENT_MAPENTRY._serialized_options = b'8\001'
  _LOCALIZER_MAPENTRY._options = None
  _LOCALIZER_MAPENTRY._serialized_options = b'8\001'
  _RELATION_MAPENTRY._options = None
  _RELATION_MAPENTRY._serialized_options = b'8\001'
  _PROPERTY_MAPENTRY._options = None
  _PROPERTY_MAPENTRY._serialized_options = b'8\001'
  _METADATA_MAPENTRY._options = None
  _METADATA_MAPENTRY._serialized_options = b'8\001'
  _SOURCE._serialized_start=4461
  _SOURCE._serialized_end=4545
  _STATE._serialized_start=4547
  _STATE._serialized_end=4596
  _FILE_TYPE._serialized_start=4599
  _FILE_TYPE._serialized_end=4769
  _OPERATION._serialized_start=106
  _OPERATION._serialized_end=150
  _FILE._serialized_start=152
  _FILE._serialized_end=262
  _ADDRESS._serialized_start=264
  _ADDRESS._serialized_end=346
  _AFFILIATION._serialized_start=348
  _AFFILIATION._serialized_end=438
  _PERSON._serialized_start=441
  _PERSON._serialized_end=656
  _EXPERIMENTALSAMPLE._serialized_start=659
  _EXPERIMENTALSAMPLE._serialized_end=870
  _EXPERIMENTALSAMPLE_MAPENTRY._serialized_start=828
  _EXPERIMENTALSAMPLE_MAPENTRY._serialized_end=870
  _EXPERIMENTALSETUP._serialized_start=873
  _EXPERIMENTALSETUP._serialized_end=1074
  _EXPERIMENTALSETUP_MAPENTRY._serialized_start=828
  _EXPERIMENTALSETUP_MAPENTRY._serialized_end=870
  _OPTICALUNIT._serialized_start=1077
  _OPTICALUNIT._serialized_end=1307
  _ILLUMINATION._serialized_start=1310
  _ILLUMINATION._serialized_end=1546
  _ILLUMINATION_MAPENTRY._serialized_start=828
  _ILLUMINATION_MAPENTRY._serialized_end=870
  _DETECTION._serialized_start=1549
  _DETECTION._serialized_end=1722
  _DETECTION_MAPENTRY._serialized_start=828
  _DETECTION_MAPENTRY._serialized_end=870
  _CAMERA._serialized_start=1725
  _CAMERA._serialized_end=2127
  _CAMERA_MAPENTRY._serialized_start=828
  _CAMERA_MAPENTRY._serialized_end=870
  _ACQUISITION._serialized_start=2130
  _ACQUISITION._serialized_end=2523
  _ACQUISITION_MAPENTRY._serialized_start=828
  _ACQUISITION_MAPENTRY._serialized_end=870
  _LIGHTSHEET._serialized_start=2526
  _LIGHTSHEET._serialized_end=2716
  _LIGHTSHEET_MAPENTRY._serialized_start=828
  _LIGHTSHEET_MAPENTRY._serialized_end=870
  _EXPERIMENT._serialized_start=2719
  _EXPERIMENT._serialized_end=2997
  _EXPERIMENT_MAPENTRY._serialized_start=828
  _EXPERIMENT_MAPENTRY._serialized_end=870
  _LOCALIZER._serialized_start=3000
  _LOCALIZER._serialized_end=3221
  _LOCALIZER_MAPENTRY._serialized_start=828
  _LOCALIZER_MAPENTRY._serialized_end=870
  _RELATION._serialized_start=3224
  _RELATION._serialized_end=3391
  _RELATION_MAPENTRY._serialized_start=828
  _RELATION_MAPENTRY._serialized_end=870
  _PROPERTY._serialized_start=3394
  _PROPERTY._serialized_end=3571
  _PROPERTY_MAPENTRY._serialized_start=828
  _PROPERTY_MAPENTRY._serialized_end=870
  _METADATA._serialized_start=3574
  _METADATA._serialized_end=4459
  _METADATA_MAPENTRY._serialized_start=828
  _METADATA_MAPENTRY._serialized_end=870
# @@protoc_insertion_point(module_scope)



"""
Utility functions for binning and rendering in 2 and 3 dimensions.

"""
import numpy as np

from locan.data.region import Rectangle, Ellipse, Polygon, RoiRegion


__all__ = []


def _napari_shape_to_region(vertices, bin_edges, region_type):
    """
    Convert napari shape to `locan.Region`.

    Parameters
    ----------
    vertices : numpy.ndarray of float
        Sequence of point coordinates as returned by napari.
    bin_edges : tuple, list, numpy.ndarray of float with shape (n_dimension, n_bin_edges)
        Array of bin edges for each dimension. At this point there are only equally-sized bins allowed.
    region_type : str
        String specifying the selector widget that can be either rectangle, ellipse, or polygon.

    Returns
    -------
    Region
    """
    # at this point there are only equally-sized bins used.
    bin_sizes = [bedges[1] - bedges[0] for bedges in bin_edges]

    vertices = np.array([bedges[0] + vert * bin_size
                         for vert, bedges, bin_size in zip(vertices.T, bin_edges, bin_sizes)]
                        ).T

    if region_type == 'rectangle':
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError('Rotated rectangles are not implemented.')
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        corner_x, corner_y = mins
        width, height = maxs - mins
        angle = 0
        region = Rectangle((corner_x, corner_y), width, height, angle)

    elif region_type == 'ellipse':
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError('Rotated ellipses are not implemented.')
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        width, height = maxs - mins
        center_x, center_y = mins[0] + width/2, mins[1] + height/2
        angle = 0
        region = Ellipse((center_x, center_y), width, height, angle)

    elif region_type == 'polygon':
        region = Polygon(np.concatenate([vertices, [vertices[0]]], axis=0))

    else:
        raise TypeError(f' Type {region_type} is not defined.')

    return region


def _napari_shape_to_RoiRegion(vertices, bin_edges, region_type):
    """
    Convert napari shape to locan RoiRegion

    Parameters
    ----------
    vertices : numpy.ndarray of float
        Sequence of point coordinates as returned by napari
    bin_edges : tuple, list, numpy.ndarray of float with shape (n_dimension, n_bin_edges)
        Array of bin edges for each dimension. At this point there are only equally-sized bins allowed.
    region_type : str
        String specifying the selector widget that can be either rectangle, ellipse, or polygon.

    Returns
    -------
    RoiRegion

    Warnings
    --------
    This function is only used by :class:`locan.RoiLegacy_0` and will be deprecated.
    Use :func:`locan.render.render2d._napari_shape_to_region` instead.
    """
    # at this point there are only equally-sized bins used.
    bin_sizes = [bedges[1] - bedges[0] for bedges in bin_edges]

    # flip since napari returns vertices with first component representing the horizontal axis
    vertices = np.flip(vertices, axis=1)

    vertices = np.array([bedges[0] + vert * bin_size
                         for vert, bedges, bin_size in zip(vertices.T, bin_edges, bin_sizes)]
                        ).T

    if region_type == 'rectangle':
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError('Rotated rectangles are not implemented.')
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        corner_x, corner_y = mins
        width, height = maxs - mins
        angle = 0
        region_specs = ((corner_x, corner_y), width, height, angle)

    elif region_type == 'ellipse':
        if len(set(vertices[:, 0].astype(int))) != 2:
            raise NotImplementedError('Rotated ellipses are not implemented.')
        mins = vertices.min(axis=0)
        maxs = vertices.max(axis=0)
        width, height = maxs - mins
        center_x, center_y = mins[0] + width/2, mins[1] + height/2
        angle = 0
        region_specs = ((center_x, center_y), width, height, angle)

    elif region_type == 'polygon':
        region_specs = np.concatenate([vertices, [vertices[0]]], axis=0)

    else:
        raise TypeError(f' Type {region_type} is not defined in locan.')

    return RoiRegion(region_specs=region_specs, region_type=region_type)



"""

Utility functions.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   system_information
   miscellaneous

"""
from .system_information import *
from .miscellaneous import *


__all__ = []
__all__.extend(system_information.__all__)
__all__.extend(miscellaneous.__all__)



"""
Compute on- and off-periods from localization frames.

Assuming that the provided localizations are acquired from the same label, we analyze the times of recording as
provided by the `frame` property.
"""
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats

from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.data.locdata import LocData


__all__ = ['BlinkStatistics']

logger = logging.getLogger(__name__)


##### The algorithms

def _blink_statistics(locdata, memory=0, remove_heading_off_periods=True):
    """
    Estimate on and off times from the frame values provided.

    On and off-periods and the first frame of each period are determined from the sorted frame values.
    A series of frame values that constantly increase by one is considered a on-period.
    Each series of missing frame values between two given frame values is considered an off-period.

    Parameters
    ----------
    locdata : LocData, array-like
        Localization data or just the frame values of given localizations.
    memory : int
        The maximum number of intermittent frames without any localization
        that are still considered to belong to the same on-period.

    Returns
    -------
    dict with numpy.ndarray as values
        'on_periods' and 'off_periods' in units of frame numbers.
        'on_periods_frame' and 'off_periods_frame' with the first frame in each on/off-period.
        'on_periods_indices' are groups of indices to the input frames or more precise np.unique(frames)
    """
    if isinstance(locdata, LocData):
        frames = locdata.data.frame.values
    else:
        frames = locdata

    frames, counts = np.unique(frames, return_counts=True)

    # provide warning if duplicate frames are found. This should not be the case for appropriate localization clusters.
    if np.any(counts > 1):
        counts_larger_one = counts[counts > 1]
        logger.warning(f'There are {sum(counts_larger_one) - len(counts_larger_one)} '
                       f'duplicated frames found that will be ignored.')

    # shift frames and add first frame if no zero frame present to account for initial off_period
    first_frame = frames[0]
    frames_ = np.insert(frames + 1, 0, 0)

    differences = np.insert(np.diff(frames_), 0, 1)  # first frame should be one since zero frame now is always on.

    # on_ and off_periods
    mask = np.where(differences > memory + 1, True, False)
    off_periods = differences[mask] - 1
    on_periods_frame = np.insert(frames_[mask], 0, 0)
    off_periods_frame = frames_[mask] - off_periods

    indices_on = np.nonzero(mask)[0]
    groups = np.split(differences, indices_on)

    # the sum is taken to include memory > 0.
    # one is added since a single localization is considered to be on for one frame.
    on_periods = np.array([np.sum(group[1:]) + 1 for group in groups])

    # grouped indices to all localizations in each on-period
    indices = np.arange(-1, len(frames_) - 1)
    on_periods_indices = np.split(indices, indices_on)

    # clean up initial shift and insert
    if first_frame == 0:  # the first frame is on
        on_periods[0] = on_periods[0] - 1
        on_periods_frame[0] = on_periods_frame[0] + 1
        on_periods_indices[0] = on_periods_indices[0][1:]

    elif first_frame > memory:  # there is an initial off_period
        on_periods = on_periods[1:]
        on_periods_frame = on_periods_frame[1:]
        if remove_heading_off_periods:
            off_periods = off_periods[1:]
            off_periods_frame = off_periods_frame[1:]
        on_periods_indices= on_periods_indices[1:]

    elif first_frame <= memory:  # there is an initial off_period integrated in the first on_period
        if remove_heading_off_periods:
            on_periods[0] = on_periods[0] - first_frame - 1
            on_periods_frame[0] = first_frame + 1
        else:
            on_periods[0] = on_periods[0] - 1
            on_periods_frame[0] = on_periods_frame[0] + 1
        on_periods_indices[0] = on_periods_indices[0][1:]

    on_periods_frame = on_periods_frame - 1
    off_periods_frame = off_periods_frame - 1

    return dict(on_periods=on_periods, on_periods_frame=on_periods_frame,
                off_periods=off_periods, off_periods_frame=off_periods_frame,
                on_periods_indices=on_periods_indices)


##### The specific analysis classes

class BlinkStatistics(_Analysis):
    """
    Estimate on and off times from the frame values provided.

    On and off-periods and the first frame of each period are determined from the sorted frame values.
    A series of frame values that constantly increase by one is considered a on-period.
    Each series of missing frame values between two given frame values is considered an off-period.

    A log warning is provided if a frame number occurs multiple times.

    Missing localizations within an on-period can be taken into account by increasing the `memory` parameter.
    There is no way to correct for false positive localizations.

    Parameters
    ----------
    memory : int
        The maximum number of intermittent frames without any localization
        that are still considered to belong to the same on-period.
    remove_heading_off_periods : bool
        Flag to indicate if off-periods at the beginning of the series are excluded.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : dict with numpy.ndarray as values
        'on_periods' and 'off_periods' in units of frame numbers.
        'on_periods_frame' and 'off_periods_frame' with the first frame in each on/off-period.
        'on_periods_indices' are groups of indices to the input frames or more precise np.unique(frames)
    """
    count = 0

    def __init__(self, meta=None, memory=0, remove_heading_off_periods=True):
        super().__init__(meta, memory=memory, remove_heading_off_periods=remove_heading_off_periods)
        self.results = None
        self.distribution_statistics = {}

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData, array-like
            Localization data or just the frame values of given localizations.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _blink_statistics(locdata=locdata, **self.parameter)
        return self

    def fit_distributions(self, distribution=stats.expon, data_identifier=('on_periods', 'off_periods'),
                          with_constraints=True, **kwargs):
        """
        Fit probability density functions to the distributions of on- and off-periods in the results
        using MLE (scipy.stats).

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        data_identifier : str
            String to identify the data in `results` for which to fit an appropriate distribution, here
            'on_periods' or 'off_periods'. For True all are fitted.
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        kwargs : dict
            Other parameters are passed to the `scipy.stat.distribution.fit()` function.
        """
        if not self:
            logger.warning('No results available to fit.')
        else:
            if isinstance(data_identifier, (tuple, list)):
                data_identifier_ = data_identifier
            else:
                data_identifier_ = (data_identifier,)

            for data_id in data_identifier_:
                self.distribution_statistics[data_id] = _DistributionFits(self, data_identifier=data_id,
                                                 distribution=distribution)
                self.distribution_statistics[data_id].fit(with_constraints=with_constraints, **kwargs)

    def hist(self, data_identifier='on_periods', ax=None, bins='auto', log=True, fit=True, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing hist(results).

        Parameters
        ----------
        data_identifier : str
            'on_periods' or 'off_periods'.
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        bins : float
            Bin specifications (passed to :func:`matplotlib.hist`).
        log : Bool
            Flag for plotting on a log scale.
        fit: bool, None
            Flag indicating if distribution fit is shown. The fit will only be computed if `distribution_statistics`
             is None.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        ax.hist(self.results[data_identifier], bins=bins, **dict(dict(density=True, log=log), **kwargs))
        ax.set(title = f'Distribution of {data_identifier}',
               xlabel = f'{data_identifier} (frames)',
               ylabel = 'PDF'
               )

        # fit distributions:
        if fit:
            if data_identifier in self.distribution_statistics and \
                    isinstance(self.distribution_statistics[data_identifier], _DistributionFits):
                self.distribution_statistics[data_identifier].plot(ax=ax)
            else:
                self.fit_distributions(data_identifier=data_identifier)
                self.distribution_statistics[data_identifier].plot(ax=ax)

        return ax


class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by specific Analysis methods.
    It holds the statistical parameters derived by fitting the result distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : Analysis object
        The analysis class with result data to fit.
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    data_identifier : str
        String to identify the data in `results` for which to fit an appropriate distribution

    Attributes
    ----------
    analyis_class : Analysis object
        The analysis class with result data to fit.
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    data_identifier : str
        String to identify the data in `results` for which to fit an appropriate distribution
    parameters : list of str
        Distribution parameters.
    """
    def __init__(self, analysis_class, distribution, data_identifier):
        self.analysis_class = analysis_class
        self.distribution = distribution
        self.data_identifier = data_identifier
        self.parameters = []

    def __repr__(self):
        """ Return representation of the _DistributionFits class. """
        param_dict = dict(analysis_class=self.analysis_class.__class__.__name__,
                          distribution=self.distribution.__class__.__name__,
                          data_identifier=self.data_identifier)
        param_string = ", ".join((f'{key}={val}' for key, val in param_dict.items()))
        return f'{self.__class__.__name__}({param_string})'

    def fit(self, with_constraints=True, **kwargs):
        """
        Fit scipy.stats.distribution to analysis_class.results[data_identifier].

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then floc=np.min(self.analysis_class.results[self.data_identifier].values).

        Parameters
        ----------
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        kwargs : dict
            Other parameters are passed to the `scipy.stat.distribution.fit()` function.
        """
        # set data
        if isinstance(self.analysis_class.results, pd.DataFrame):
            data = self.analysis_class.results[self.data_identifier].values
        else:
            data = self.analysis_class.results[self.data_identifier]

        # define parameter names
        for param in _list_parameters(self.distribution):
            self.parameters.append(self.data_identifier + '_' + param)

        # perform fit
        if with_constraints and self.distribution == stats.expon:
            # MLE fit of exponential distribution with constraints
            fit_results = stats.expon.fit(data, **dict(dict(floc=np.min(data)), **kwargs))
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)
        else:
            fit_results = self.distribution.fit(data, **kwargs)
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)

    def plot(self, ax=None, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the probability distribution functions of fitted
        results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        parameter = self.parameter_dict().values()
        x_values = np.linspace(self.distribution.ppf(1e-4, *parameter),
                               self.distribution.ppf(1 - 1e-4, *parameter), 100)
        ax.plot(x_values, self.distribution.pdf(x_values, *parameter), 'r-',
                **dict(dict(lw=3, alpha=0.6, label=str(self.distribution.name) + ' pdf'),
                       **kwargs))

        return ax

    def parameter_dict(self):
        """ Dictionary of fitted parameters. """
        return {k: self.__dict__[k] for k in self.parameters}



"""

A class to carry localization data.

"""
import time
import warnings
from itertools import accumulate
import copy
import logging

from google.protobuf import text_format, json_format
import numpy as np
import pandas as pd
try:
    from scipy.spatial import QhullError
except ImportError:
    from scipy.spatial.qhull import QhullError # needed for Python 3.7

from locan import locdata_id  # is required to use locdata_id as global variable
from locan.constants import PROPERTY_KEYS, PropertyKey
from locan.data.region import Region, RoiRegion
import locan.data.hulls
from locan.data import metadata_pb2
from locan.data.metadata_utils import _modify_meta, metadata_to_formatted_string
from locan.data.properties import inertia_moments


__all__ = ['LocData']

logger = logging.getLogger(__name__)


class LocData:
    """
    This class carries localization data, aggregated properties and meta data.

    Data consist of individual elements being either localizations or other `LocData` objects. Both, localizations and
    `Locdata` objects have properties. Properties come from the original data or are added by analysis procedures.

    Parameters
    ----------
    references : LocData, list(LocData), None
        A `LocData` reference or an array with references to `LocData` objects referring to the selected localizations
        in dataset.
    dataframe : pandas.DataFrame, None
        Dataframe with localization data.
    indices : slice object, list(int), None
        Indices for dataframe in references that makes up the data. `indices` refers to index label, not position.
    meta : locan.data.metadata_pb2.Metadata, dictionary
        Metadata about the current dataset and its history.

    Attributes
    ----------
    references : LocData, list(LocData), None
        A LocData reference or an array with references to LocData objects referring to the selected localizations
        in dataframe.
    dataframe : pandas.DataFrame, None
        Dataframe with localization data.
    indices : slice object, list(int), None
        Indices for dataframe in references that makes up the data.
    meta : locan.data.metadata_pb2.Metadata
        Metadata about the current dataset and its history.
    properties : pandas.DataFrame
        List of properties generated from data.
    coordinate_labels : list of str
        The available coordinate properties.
    dimension : int
        Number of coordinates available for each localization (i.e. size of `coordinate_labels`).
    """
    count = 0
    """int: A counter for counting LocData instantiations (class attribute)."""

    def __init__(self, references=None, dataframe=pd.DataFrame(), indices=None,
                 meta=None):
        self.__class__.count += 1

        self.references = references
        self.dataframe = dataframe
        self.indices = indices
        self.meta = metadata_pb2.Metadata()
        self.properties = {}

        # regions and hulls
        self._region = None
        self._bounding_box = None
        self._oriented_bounding_box = None
        self._convex_hull = None
        self._alpha_shape = None
        self._inertia_moments = None

        self.coordinate_labels = sorted(list(set(self.data.columns).intersection({'position_x',
                                                                                  'position_y',
                                                                                  'position_z'})))

        self.dimension = len(self.coordinate_labels)

        self._update_properties()

        # meta
        global locdata_id
        locdata_id += 1
        self.meta.identifier = str(locdata_id)

        self.meta.creation_time.GetCurrentTime()
        self.meta.source = metadata_pb2.DESIGN
        self.meta.state = metadata_pb2.RAW

        self.meta.element_count = len(self.data.index)
        if 'frame' in self.data.columns:
            self.meta.frame_count = len(self.data['frame'].unique())

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(self.meta, key, value)
        else:
            self.meta.MergeFrom(meta)

    def _update_properties(self):
        self.properties['localization_count'] = len(self.data.index)

        # property for mean spatial coordinates (centroids)
        self.properties.update(dict(self.data[self.coordinate_labels].mean()))

        self.bounding_box  # update self._bounding_box

    def __del__(self):
        """Updating the counter upon deletion of class instance."""
        self.__class__.count -= 1

    def __len__(self):
        """Return the length of data, i.e. the number of elements (localizations or collection elements)."""
        return len(self.data.index)

    def __getstate__(self):
        """Modify pickling behavior."""
        # Copy the object's state from self.__dict__ to avoid modifying the original state.
        state = self.__dict__.copy()
        # Serialize the unpicklable protobuf entries.
        json_string = json_format.MessageToJson(self.meta, including_default_value_fields=False)
        state['meta'] = json_string
        return state

    def __setstate__(self, state):
        """Modify pickling behavior."""
        # Restore instance attributes.
        self.__dict__.update(state)
        # Restore protobuf class for meta attribute
        self.meta = metadata_pb2.Metadata()
        self.meta = json_format.Parse(state['meta'], self.meta)

    def __copy__(self):
        """
        Create a shallow copy of locdata (keeping all references) with the following exceptions:
        (i) The class variable `count` is increased for the copied LocData object.
        (ii) Metadata keeps the original meta.creation_time while meta.modification_time and meta.history is updated.
        """
        new_locdata = LocData(self.references,
                              self.dataframe,
                              self.indices,
                              meta=None)
        new_locdata._region = self._region
        # meta
        meta_ = _modify_meta(self, new_locdata, function_name='LocData.copy',
                             parameter=None, meta=None)
        new_locdata.meta = meta_
        return new_locdata

    def __deepcopy__(self, memodict=None):
        """
        Create a deep copy of locdata (including all references) with the following exceptions:
        (i) The class variable `count` is increased for all deepcopied LocData objects.
        (ii) Metadata keeps the original meta.creation_time while meta.modification_time and meta.history is updated.
        """
        if memodict is None:
            memodict = {}
        new_locdata = LocData(copy.deepcopy(self.references, memodict),
                              copy.deepcopy(self.dataframe, memodict),
                              copy.deepcopy(self.indices, memodict),
                              meta=None)
        new_locdata._region = self._region
        # meta
        meta_ = _modify_meta(self, new_locdata, function_name='LocData.deepcopy',
                             parameter=None, meta=None)
        new_locdata.meta = meta_
        return new_locdata

    @property
    def bounding_box(self):
        """Hull object: Return an object representing the axis-aligned minimal bounding box."""
        if self._bounding_box is None:
            try:
                self._bounding_box = locan.data.hulls.BoundingBox(self.coordinates)
                self.properties['region_measure_bb'] = self._bounding_box.region_measure
                if self._bounding_box.region_measure:
                    self.properties['localization_density_bb'] = \
                        self.properties['localization_count'] / self._bounding_box.region_measure
                if self._bounding_box.subregion_measure:
                    self.properties['subregion_measure_bb'] = self._bounding_box.subregion_measure
            except ValueError:
                warnings.warn('Properties related to bounding box could not be computed.', UserWarning)
        return self._bounding_box

    @property
    def convex_hull(self):
        """Hull object: Return an object representing the convex hull of all localizations."""
        if self._convex_hull is None:
            try:
                self._convex_hull = locan.data.hulls.ConvexHull(self.coordinates)
                self.properties['region_measure_ch'] = self._convex_hull.region_measure
                if self._convex_hull.region_measure:
                    self.properties['localization_density_ch'] = self.properties['localization_count'] \
                                                                      / self._convex_hull.region_measure
            except (TypeError, QhullError):
                warnings.warn('Properties related to convex hull could not be computed.', UserWarning)
        return self._convex_hull

    @property
    def oriented_bounding_box(self):
        """Hull object: Return an object representing the oriented minimal bounding box."""
        if self._oriented_bounding_box is None:
            try:
                self._oriented_bounding_box = locan.data.hulls.OrientedBoundingBox(self.coordinates)
                self.properties['region_measure_obb'] = self._oriented_bounding_box.region_measure
                if self._oriented_bounding_box.region_measure:
                    self.properties['localization_density_obb'] = self.properties['localization_count'] \
                                                                      / self._oriented_bounding_box.region_measure
                self.properties['orientation_obb'] = self._oriented_bounding_box.angle
                self.properties['circularity_obb'] = self._oriented_bounding_box.elongation
            except TypeError:
                warnings.warn('Properties related to oriented bounding box could not be computed.', UserWarning)
        return self._oriented_bounding_box

    @property
    def alpha_shape(self):
        """Hull object: Return an object representing the alpha-shape of all localizations."""
        return self._alpha_shape

    def update_alpha_shape(self, alpha):
        """Compute the alpha shape for specific `alpha` and update `self.alpha_shape`.

        Parameters
        ----------
        alpha : float
            Alpha parameter specifying a unique alpha complex.

        Returns
        -------
        LocData
            The modified object
        """
        try:
            if self._alpha_shape is None:
                self._alpha_shape = locan.data.hulls.AlphaShape(points=self.coordinates, alpha=alpha)
            else:
                self._alpha_shape.alpha = alpha

            self.properties['region_measure_as'] = self._alpha_shape.region_measure
            try:
                self.properties['localization_density_as'] = self._alpha_shape.n_points_alpha_shape \
                                                             / self._alpha_shape.region_measure
            except ZeroDivisionError:
                self.properties['localization_density_as'] = float('nan')

        except TypeError:
            warnings.warn('Properties related to alpha shape could not be computed.', UserWarning)
        return self

    def update_alpha_shape_in_references(self, alpha):
        """
        Compute the alpha shape for each element in `locdata.references` and update `locdata.dataframe`.

        Returns
        -------
        LocData
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.update_alpha_shape(alpha=alpha)
            new_df = pd.DataFrame([reference.properties for reference in self.references])
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [column for column in new_df.columns if column in self.dataframe.columns]
            new_df.drop(columns=new_columns, inplace=True, errors='ignore')
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    @property
    def inertia_moments(self):
        """Inertia moments are returned as computed by :func:`locan.data.properties.inertia_moments`."""
        if self._inertia_moments is None:
            try:
                self._inertia_moments = locan.data.properties.inertia_moments(self.coordinates)
                self.properties['orientation_im'] = self._inertia_moments.orientation
                self.properties['circularity_im'] = self._inertia_moments.eccentricity
            except TypeError:
                warnings.warn('Properties related to inertia_moments could not be computed.', UserWarning)
        return self._inertia_moments

    def update_inertia_moments_in_references(self):
        """
        Compute inertia_moments for each element in locdata.references and update locdata.dataframe.

        Returns
        -------
        LocData
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.inertia_moments  # request property to update
            new_df = pd.DataFrame([reference.properties for reference in self.references])
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [column for column in new_df.columns if column in self.dataframe.columns]
            new_df.drop(columns=new_columns, inplace=True, errors='ignore')
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    @property
    def region(self):
        """RoiRegion object: Return the region that supports all localizations."""
        return self._region

    @region.setter
    def region(self, region):
        if region is not None:
            if region.dimension != self.dimension:
                raise TypeError("Region dimension and coordinates dimension must be identical.")
            elif len(self) != len(region.contains(self.coordinates)):
                logger.warning("Not all coordinates are within region.")

        if isinstance(region, (Region, RoiRegion)) or region is None:
            self._region = region

        elif isinstance(region, dict):  # legacy code to deal with deprecated RoiLegacy_0
            region_ = RoiRegion(**region)
            if region_ is not None:
                if region_.dimension != self.dimension:
                    raise TypeError("Region dimension and coordinates dimension must be identical.")
                elif len(self) != len(region_.contains(self.coordinates)):
                    logger.warning("Not all coordinates are within region.")
            self._region = region_

        else:
            raise TypeError

        # property for region measures
        if self._region is not None:
            if self._region.region_measure:
                self.properties['region_measure'] = self._region.region_measure
                self.properties['localization_density'] = self.meta.element_count / self._region.region_measure
            if self._region.subregion_measure:
                self.properties['subregion_measure'] = self._region.subregion_measure

    @property
    def data(self):
        """pandas.DataFrame: Return all elements either copied from the reference or referencing the current
        dataframe. """
        if isinstance(self.references, LocData):
            # we refer to the localization data by its index label, not position
            # in other words we decided not to use iloc but loc
            # df = self.references.data.loc[self.indices]  ... but this does not work in pandas.
            # also see:
            # https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
            try:
                df = self.references.data.loc[self.indices]
            except KeyError:
                df = self.references.data.loc[self.references.data.index.intersection(self.indices)]
            df = pd.merge(df, self.dataframe, left_index=True, right_index=True, how='outer')
            return df
        else:
            return self.dataframe

    @property
    def coordinates(self):
        """ndarray: Return all coordinate values. """
        return self.data[self.coordinate_labels].values

    @property
    def centroid(self):
        """ndarray: Return coordinate values of the centroid
        (being the property values for all coordinate labels)."""
        return np.array([self.properties[coordinate_label] for coordinate_label in self.coordinate_labels])

    @classmethod
    def from_dataframe(cls, dataframe=pd.DataFrame(), meta=None):
        """
        Create new LocData object from pandas.DataFrame with localization data.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            Localization data.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """
        dataframe = dataframe
        meta_ = metadata_pb2.Metadata()

        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.history.add(name='LocData.from_dataframe')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        return cls(dataframe=dataframe, meta=meta_)

    @classmethod
    def from_coordinates(cls, coordinates=(), coordinate_labels=None, meta=None):
        """
        Create new LocData object from a sequence of localization coordinates.

        Parameters
        ----------
        coordinates : sequence of tuples with shape (n_loclizations, dimension)
            Sequence of tuples with localization coordinates
        coordinate_labels : sequence of str
            The available coordinate properties.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """
        if np.size(coordinates):
            dimension = len(coordinates[0])

            if coordinate_labels is None:
                coordinate_labels = ['position_x', 'position_y', 'position_z'][0:dimension]
            else:
                if all(cl in PROPERTY_KEYS for cl in coordinate_labels):
                    coordinate_labels = coordinate_labels
                else:
                    raise ValueError('The given coordinate_labels are not standard property keys.')

            dataframe = pd.DataFrame.from_records(data=coordinates, columns=coordinate_labels)

        else:
            dataframe = pd.DataFrame()

        meta_ = metadata_pb2.Metadata()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.history.add(name='LocData.from_coordinates')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        return cls(dataframe=dataframe, meta=meta_)

    @classmethod
    def from_selection(cls, locdata, indices=slice(0, None), meta=None):
        """
        Create new LocData object from selected elements in another `LocData`.

        Parameters
        ----------
        locdata : LocData
            Locdata object from which to select elements.
        indices : slice object, list(int), None
            Index labels for elements in locdata that make up the new data.
            Note that contrary to usual python slices, both the start and the stop are included
            (see pandas documentation). `Indices` refer to index value not position in list.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the selected data.

        Note
        ----
        No error is raised if indices do not exist in locdata.
        """
        references = locdata
        indices = indices
        meta_ = metadata_pb2.Metadata()
        meta_.CopyFrom(locdata.meta)
        try:
            meta_.ClearField("identifier")
        except ValueError:
            pass

        try:
            meta_.ClearField("element_count")
        except ValueError:
            pass

        try:
            meta_.ClearField("frame_count")
        except ValueError:
            pass

        meta_.modification_time.GetCurrentTime()
        meta_.state = metadata_pb2.MODIFIED
        meta_.ancestor_identifiers.append(locdata.meta.identifier)
        meta_.history.add(name='LocData.from_selection')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        new_locdata = cls(references=references, indices=indices, meta=meta_)
        new_locdata.region = references.region
        return new_locdata

    @classmethod
    def from_collection(cls, locdatas, meta=None):
        """
        Create new LocData object by collecting LocData objects.

        Parameters
        ----------
        locdatas : list of LocData
            Locdata objects to collect.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """
        references = locdatas
        dataframe = pd.DataFrame([ref.properties for ref in references])

        meta_ = metadata_pb2.Metadata()

        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.ancestor_identifiers[:] = [ref.meta.identifier for ref in references]
        meta_.history.add(name='LocData.from_collection')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        return cls(references=references, dataframe=dataframe, meta=meta_)

    @classmethod
    def concat(cls, locdatas, meta=None):
        """
        Concatenate LocData objects.

        Parameters
        ----------
        locdatas : list of LocData
            Locdata objects to concatenate.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with dataframe representing the concatenated data.
        """

        dataframe = pd.concat([i.data for i in locdatas], ignore_index=True, sort=False)

        # concatenate references also if None
        references = []
        for locdata in locdatas:
            try:
                references.extend(locdata.references)
            except TypeError:
                references.append(locdata.references)

        # check if all elements are None
        if not any(references):
            references = None

        meta_ = metadata_pb2.Metadata()

        meta_.creation_time.GetCurrentTime()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.MODIFIED
        meta_.ancestor_identifiers[:] = [dat.meta.identifier for dat in locdatas]
        meta_.history.add(name='concat')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        return cls(references=references, dataframe=dataframe, meta=meta_)

    @classmethod
    def from_chunks(cls, locdata, chunks=None, chunk_size=None, n_chunks=None, order='successive', drop=False, meta=None):
        """
        Divide locdata in chunks of localization elements.

        Parameters
        ----------
        locdatas : list of LocData
            Locdata objects to concatenate.
        chunks : list[tuples]
            Localization chunks as defined by a list of index-tuples
        chunk_size : int, None
            Number of localizations per chunk. One of `chunk_size` or `n_chunks` must be different from None.
        n_chunks : int, None
            Number of chunks. One of `chunk_size` or `n_chunks` must be different from None.
        order : str
            The order in which to select localizations. One of 'successive' or 'alternating'.
        drop : bool
            If True the last chunk will be eliminated if it has fewer localizations than the other chunks.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            A new LocData instance with references and dataframe elements representing the individual chunks.
        """
        n_nones = sum(element is None for element in [chunks, chunk_size, n_chunks])

        if n_nones != 2:
            raise ValueError("One and only one of `chunks`, `chunk_size` or `n_chunks` must be different from None.")
        elif chunks is not None:
            index_lists = list(chunks)
        else:
            if chunk_size is not None:
                if (len(locdata) % chunk_size) == 0:
                    n_chunks = len(locdata) // chunk_size
                else:
                    n_chunks = len(locdata) // chunk_size + 1
            else:  # if n_chunks is not None
                if (len(locdata) % n_chunks) == 0:
                    chunk_size = len(locdata) // n_chunks
                else:
                    chunk_size = len(locdata) // (n_chunks-1)

            if order == 'successive':
                if (len(locdata) % chunk_size) == 0:
                    chunk_sizes = [chunk_size] * n_chunks
                else:
                    chunk_sizes = [chunk_size] * (n_chunks-1) + [(len(locdata) % chunk_size)]
                cum_chunk_sizes = list(accumulate(chunk_sizes))
                cum_chunk_sizes.insert(0, 0)
                index_lists = [locdata.data.index[slice(lower, upper)]
                               for lower, upper in zip(cum_chunk_sizes[:-1], cum_chunk_sizes[1:])]

            elif order == 'alternating':
                index_lists = [locdata.data.index[slice(i_chunk, None, n_chunks)] for i_chunk in range(n_chunks)]

            else:
                raise ValueError(f"The order {order} is not implemented.")

        if drop and len(index_lists) > 1 and len(index_lists[-1]) < len(index_lists[0]):
            index_lists = index_lists[:-1]

        references = [LocData.from_selection(locdata=locdata, indices=index_list) for index_list in index_lists]
        dataframe = pd.DataFrame([ref.properties for ref in references])

        meta_ = metadata_pb2.Metadata()

        meta_.creation_time.GetCurrentTime()
        meta_.source = metadata_pb2.DESIGN
        meta_.state = metadata_pb2.RAW
        meta_.ancestor_identifiers[:] = [ref.meta.identifier for ref in references]
        meta_.history.add(name='LocData.chunks')

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(meta_, key, value)
        else:
            meta_.MergeFrom(meta)

        return cls(references=references, dataframe=dataframe, meta=meta_)

    def reset(self, reset_index=False):
        """
        Reset hulls and properties. This is needed after the dataframe attribute has been modified in place.

        Note
        ----
        Should be used with care because metadata is not updated accordingly.
        The region property is not changed.
        Better to just re-instantiate with `LocData.from_dataframe()` or use `locdata.update()`.

        Parameters
        ----------
        reset_index : Bool
            Flag indicating if the index is reset to integer values. If True the previous index values are discarded.

        Returns
        -------
        LocData
            The modified object
        """
        if reset_index is True:
            self.dataframe.reset_index(drop=True, inplace=True)

        self.properties = {}
        self._bounding_box = None
        self._oriented_bounding_box = None
        self._convex_hull = None
        self._alpha_shape = None

        self._update_properties()

        return self

    def update(self, dataframe, reset_index=False, meta=None):
        """
        Update the dataframe attribute in place.

        Use this function rather than setting locdata.dataframe directly in order to automatically update
        the attributes for dimension, coordinate_labels, hulls, properties, and metadata.

        Parameters
        ----------
        dataframe : pandas.DataFrame, None
            Dataframe with localization data.
        reset_index : Bool
            Flag indicating if the index is reset to integer values. If True the previous index values are discarded.
        meta : locan.data.metadata_pb2.Metadata
            Metadata about the current dataset and its history.

        Returns
        -------
        LocData
            The modified object
        """
        local_parameter = locals()
        del local_parameter['dataframe']  # dataframe is obvious and possibly large and should not be repeated in meta.

        if self.references is not None:
            self.reduce(reset_index=reset_index)
            logger.warning("LocData.reduce() was applied since self.references was not None.")

        self.dataframe = dataframe
        self.coordinate_labels = sorted(list(set(self.data.columns).intersection({'position_x',
                                                                                  'position_y',
                                                                                  'position_z'})))
        self.dimension = len(self.coordinate_labels)
        self.reset(reset_index=reset_index)  # update hulls and properties

        # update meta
        self.meta.modification_time.GetCurrentTime()
        self.meta.state = metadata_pb2.MODIFIED
        self.meta.history.add(name='LocData.update', parameter=str(local_parameter))

        self.meta.element_count = len(self.data.index)
        if 'frame' in self.data.columns:
            self.meta.frame_count = len(self.data['frame'].unique())

        if meta is None:
            pass
        elif isinstance(meta, dict):
            for key, value in meta.items():
                setattr(self.meta, key, value)
        else:
            self.meta.MergeFrom(meta)

        return self

    def reduce(self, reset_index=False):
        """
        Clean up references.

        This includes to update `Locdata.dataframe` and set `LocData.references` and `LocData.indices` to None.

        Parameters
        ----------
        reset_index : Bool
            Flag indicating if the index is reset to integer values. If True the previous index values are discarded.

        Returns
        -------
        LocData
            The modified object
        """
        if self.references is None:
            pass
        elif isinstance(self.references, (LocData, list)):
            self.dataframe = self.data
            self.indices = None
            self.references = None
        else:
            raise ValueError('references has undefined value.')

        if reset_index is True:
            self.dataframe.reset_index(drop=True, inplace=True)

        return self

    def update_convex_hulls_in_references(self):
        """
        Compute the convex hull for each element in locdata.references and update locdata.dataframe.

        Returns
        -------
        LocData
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.convex_hull  # request property to update reference._convex_hull

            new_df = pd.DataFrame([reference.properties for reference in self.references])
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [column for column in new_df.columns if column in self.dataframe.columns]
            new_df.drop(columns=new_columns, inplace=True, errors='ignore')
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    def update_oriented_bounding_box_in_references(self):
        """
        Compute the oriented bounding box for each element in locdata.references and update locdata.dataframe.

        Returns
        -------
        LocData
            The modified object
        """
        if isinstance(self.references, list):
            for reference in self.references:
                reference.oriented_bounding_box  # request property to update reference._convex_hull
            new_df = pd.DataFrame([reference.properties for reference in self.references])
            new_df.index = self.data.index
            self.dataframe.update(new_df)
            new_columns = [column for column in new_df.columns if column in self.dataframe.columns]
            new_df.drop(columns=new_columns, inplace=True, errors='ignore')
            self.dataframe = pd.concat([self.dataframe, new_df], axis=1)
        return self

    def projection(self, coordinate_labels):
        """
        Reduce dimensions by projecting all localization coordinates onto selected coordinates.

        Parameters
        ----------
        coordinate_labels : str, list
            The coordinate labels to project onto.

        Returns
        -------
            LocData
        """
        local_parameter = locals()

        if isinstance(coordinate_labels, str):
            coordinate_labels = [coordinate_labels]

        new_locdata = copy.deepcopy(self)

        # reduce coordinate dimensions
        coordinate_labels_to_drop = [label for label in self.coordinate_labels if label not in coordinate_labels]
        columns = self.data.columns
        new_columns = [column for column in columns if column not in coordinate_labels_to_drop]
        dataframe = new_locdata.data[new_columns]

        # update
        _meta = metadata_pb2.Metadata()
        _meta.history.add(name='LocData.projection', parameter=str(local_parameter))
        # other updates are done in the coming update call.

        new_locdata = new_locdata.update(dataframe=dataframe, meta=_meta)

        return new_locdata

    def print_meta(self):
        """
        Print Locdata.metadata.

        See Also
        --------
        :func:`locan.data.metadata_utils.metadata_to_formatted_string`
        """
        print(metadata_to_formatted_string(self.meta))

    def print_summary(self):
        """
        Print a summary containing the most common metadata keys.
        """
        meta_ = metadata_pb2.Metadata()
        if self.meta.HasField("file"):
            meta_.file.CopyFrom(self.meta.file)
        meta_.identifier = self.meta.identifier
        meta_.comment = self.meta.comment
        meta_.creation_time.CopyFrom(self.meta.creation_time)
        if self.meta.HasField("modification_time"):
            meta_.modification_time.CopyFrom(self.meta.modification_time)
        meta_.source = self.meta.source
        meta_.state = self.meta.state
        meta_.element_count = self.meta.element_count
        meta_.frame_count = self.meta.frame_count

        print(metadata_to_formatted_string(meta_))



#!/usr/bin/env python

"""
Show system information and dependency versions.


To run the script::

    locan show_versions -v -e -o <module name> [<module name>...]

Try for instance::

    locan show_versions -v -e

See Also
--------
locan.utils.system_information.show_versions
"""
import argparse

from locan.utils.system_information import show_versions as sc_show_versions


def _add_arguments(parser):
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Include information on node and executable path.')
    parser.add_argument('-e', '--extra', dest='extra', action='store_true',
                        help='Include extra dependencies as specified in setup.py.')
    parser.add_argument('-o', '--other', dest='other', type=str, nargs='*',
                        help='Include other module names.')


def main(args=None):
    parser = argparse.ArgumentParser(description='Show system information and dependency versions.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)
    sc_show_versions(verbose=returned_args.verbose, extra_dependencies=returned_args.extra,
                     other_dependencies=returned_args.other)


if __name__ == '__main__':
    main()



"""

Utility functions for working with regions.

See Also
--------
:func:`locan.data.filter.select_by_region`
:func:`locan.data.properties.misc.distance_to_region`
:func:`locan.data.properties.misc.distance_to_region_boundary`
"""
from shapely.ops import unary_union

from locan.data.region import Region, Region2D, EmptyRegion, RoiRegion


__all__ = ['regions_union', 'expand_region', 'surrounding_region']


def regions_union(regions):
    """
    Return the union of `regions`.

    Parameters
    ----------
    regions : list of Region
        Original region(s)

    Returns
    --------
    Region
    """
    if all([isinstance(region, (Region2D, RoiRegion)) for region in regions]):
        shapely_objects = [reg.shapely_object for reg in regions]
        unified_regions = unary_union(shapely_objects)
        if unified_regions.is_empty:
            return EmptyRegion()
        else:
            return Region2D.from_shapely(unified_regions)
    else:
        raise NotImplementedError("regions must all be Region2D")


def expand_region(region, distance=100, support=None, **kwargs):
    """
    Expand a region by `distance`.
    If region contains a list of regions, the unification of all expanded regions is returned.

    Parameters
    ----------
    region : Region, shapely.Polygon
        Original region(s)
    distance : int, float
        Distance by which the region is expanded orthogonal to its boundary.
    support : Region or None
        A region defining the maximum outer boundary.
    kwargs : dict
        Other parameters passed to :func:`shapely.geometry.buffer` for :class:`Region2D` objects.

    Returns
    --------
    Region
    """
    expanded_region = region.buffer(distance, **kwargs)

    if support is not None:
        expanded_region = support.intersection(expanded_region)

    try:
        return Region2D.from_shapely(expanded_region)
    except:
        return expanded_region


def surrounding_region(region, distance=100, support=None, **kwargs):
    """
    Define surrounding region by extending a region and returning the extended region excluding the input region.
    If region contains a list of regions, the unification of all extended regions is returned.

    Parameters
    ----------
    region : Region
        Original region(s)
    distance : int, float
        Distance by which the region is extended orthogonal to its boundary.
    support : Region or None
        A region defining the maximum outer boundary.
    kwargs : dict
        Other parameters passed to :func:`shapely.geometry.buffer` for :class:`Region2D` objects.

    Returns
    --------
    Region
    """
    extended_region = expand_region(region, distance=distance, support=support, **kwargs)
    if isinstance(extended_region, Region2D):
        surrounding_region_ = extended_region.symmetric_difference(region)
        return Region2D.from_shapely(surrounding_region_)
    else:
        raise NotImplementedError("Only 2-dimensional function has been implemented.")



import time

from locan.utils.format import _time_string


def test_time_string():
    time_ = 1000000000  # result from time.time()
    string_ = _time_string(time_)
    # assert string_ == '2001-09-09 03:46:40 +0200'  # breaks in different time zones
    assert string_.startswith("2001-09-09 ")
    assert ":46:40 " in string_



"""

File input/output functions.

This module provides functions for file input and output of data related to single-molecule localization microscopy.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   locdata

"""
from .locdata import *

__all__ = []
__all__.extend(locdata.__all__)
